import sys
import numpy as np

sys.path.append("../tools")
from roggia import ElementType, TwoPortElement

class EqnUtility:

    def __init__(self, ckt):
        self.ckt = ckt

        self.sys_cap_edge_list = self.ckt.get_sys_var_cap_list()
        self.sys_ind_edge_list = self.ckt.get_sys_var_ind_list()
        self.co_tree_cap_list  = self.ckt.get_non_sys_var_cap_list()
        self.spn_tree_ind_list = self.ckt.get_non_sys_var_ind_list()
        self.co_tree_res_list  = self.ckt.get_co_tree_res_list()
        self.spn_tree_res_list = self.ckt.get_spanning_tree_res_list()
        self.dep_src_edge_list = self.ckt.get_dep_var_list()
        self.src_edge_list     = self.ckt.get_src_edge_list()

        self.num_sys_var_cap   = self.ckt.get_num_sys_var_cap()
        self.num_sys_var_ind   = self.ckt.get_num_sys_var_ind()
        self.num_co_tree_cap   = self.ckt.get_num_non_sys_var_cap()
        self.num_spn_tree_ind  = self.ckt.get_num_non_sys_var_ind()
        self.num_edges         = self.ckt.get_num_edges()
        self.num_dep_srcs      = self.ckt.get_num_dep_src()

        self.num_spn_tree_res  = len(self.spn_tree_res_list)
        self.num_co_tree_res   = len(self.co_tree_res_list)
        self.num_src_edges     = len(self.src_edge_list)

        self.qf                = self.ckt.qf.copy()
        self.bf                = self.ckt.bf.copy()

        self.cap_value         = np.zeros(self.num_edges)
        self.ind_value         = np.zeros(self.num_edges)

        self.xr                = np.ones(self.num_edges + self.num_dep_srcs)

        # spanning-tree resistors
        for i in range(self.num_spn_tree_res):
            edge_ref    = self.spn_tree_res_list[i]['ref']
            edge_info   = self.spn_tree_res_list[i]['info']

            self.xr[edge_ref] = 1 / edge_info.get_value(0, 0, 0)

        # co-tree resistors
        for i in range(self.num_co_tree_res):
            edge_ref    = self.co_tree_res_list[i]['ref']
            edge_info   = self.co_tree_res_list[i]['info']

            self.xr[edge_ref] = 1 / edge_info.get_value(0, 0, 0)

        self.dy_eqn_offset = self.num_edges + self.num_dep_srcs

        self.t_prev = 0

    def set_prev_eqn_soln(self, eqn_soln):
        self.prev_eqn_soln = eqn_soln

    def get_prev_eqn_soln(self):
        return self.prev_eqn_soln

    def set_tprev(self, t_prev):
        self.t_prev = t_prev

    def add_subckt(self, subckt):
        self.subckt = subckt

    # Component         |         Spanning-tree          |            Co-tree             |
    # ------------------+----------------+---------------+----------------+---------------+
    #                   |    Eqn Var     |    Sys Var    |    Eqn Var     |    Sys Var    |
    # ------------------+----------------+---------------+----------------+---------------+
    # Capacitor         |      dV        |     sys[x]    |      V, dV     |       -       |
    #                   |                |               |                |               |
    # Inductor          |      I, dI     |       -       |       dI       |     sys[x]    |
    #                   |                |               |                |               |
    # Resistor          |       V        |       -       |       V        |       -       |
    #                   |                |               |                |               |
    # Indep Vsrc        |       I        |       -       |       -        |       -       |
    #                   |                |               |                |               |
    # CCVS              |      V, I      |       -       |       -        |       -       |
    def dypc_eqn_nonlinear(self, x, sys_var, t):

        edge_voltages_x     = np.zeros(self.num_edges)
        edge_currents_x     = np.zeros(self.num_edges)

        dy_edge_voltages_x  = np.zeros(self.num_edges)
        dy_edge_currents_x  = np.zeros(self.num_edges)

        self.co_tree_cap_dy  = np.zeros(self.num_edges)
        self.spn_tree_ind_dy = np.zeros(self.num_edges)

        qf_iter = self.ckt.qf.copy()
        bf_iter = self.ckt.bf.copy()

        # ----------------------------------
        # Update edge voltages and currents
        # ----------------------------------
        for i in range(self.num_edges):
            edge_voltages_x[i] = x[i]
            edge_currents_x[i] = x[i]

        # voltage/current sources
        for i in range(self.num_src_edges):
            edge_ref    = self.src_edge_list[i]['ref']
            edge_info   = self.src_edge_list[i]['info']

            dependent_edge    = edge_info.get_dependant()

            if edge_info.get_type() == ElementType.voltage_src:
                if dependent_edge:
                    dep_var_ref = self.src_edge_list[i]['dep_ref']
                    edge_voltages_x[edge_ref]    = x[dep_var_ref]
                    dy_edge_voltages_x[edge_ref] = 0 # REVISIT will need another variable
                else:
                    edge_voltages_x[edge_ref]    = edge_info.get_value(0, t, 0)
                    dy_edge_voltages_x[edge_ref] = edge_info.get_dy_value(0, t, 0)
            else:
                if dependent_edge:
                    dep_var_ref = self.src_edge_list[i]['dep_ref']
                    edge_currents_x[edge_ref]    = x[dep_var_ref]
                    dy_edge_currents_x[edge_ref] = 0 # REVISIT will need another variable
                else:
                    edge_currents_x[edge_ref]    = edge_info.get_value(0, t, 0)
                    dy_edge_currents_x[edge_ref] = edge_info.get_dy_value(0, t, 0)

        # Over-ride for sys var caps
        for i in range(self.num_sys_var_cap):
            edge_ref    = self.sys_cap_edge_list[i]['ref']
            sys_var_ref = self.sys_cap_edge_list[i]['sys_var_ref']

            edge_voltages_x[edge_ref]    = sys_var[sys_var_ref]
            edge_currents_x[edge_ref]    = x[edge_ref]

        # Co-tree capacitors
        for i in range(self.num_co_tree_cap):
            edge_ref    = self.co_tree_cap_list[i]['ref']

            # Determine dVc by using KVL equation and differentiating.
            # For example, if Vc3  =  Vc1  + Vc2  - Vs then
            #                 dVc3 = dVc1 + dVc2 - dVs

            # Extract the KVL equation relating to the co-tree capacitor
            bf_row           = self.ckt.bf[edge_ref,:].copy()

            # Mask references to the co-tree capacitor in the bf row
            bf_row[edge_ref] = 0

            # Multiply by dVc of spanning-tree capacitors
            bf_x             = np.transpose(edge_currents_x.copy())

            # voltage/current sources (replace with derivatives)
            for i in range(self.num_src_edges):
                src_edge_ref       = self.src_edge_list[i]['ref']
                bf_x[src_edge_ref] = dy_edge_voltages_x[src_edge_ref]

            bf_sol = -1 * bf_row.dot(bf_x)
            self.co_tree_cap_dy[edge_ref] = bf_sol

            edge_voltages_x[edge_ref]    = x[edge_ref]
            edge_currents_x[edge_ref]    = bf_sol

        # Over-ride for sys var inductors
        for i in range(self.num_sys_var_ind):
            edge_ref    = self.sys_ind_edge_list[i]['ref']
            sys_var_ref = self.sys_ind_edge_list[i]['sys_var_ref']

            edge_currents_x[edge_ref] = sys_var[sys_var_ref + self.num_sys_var_cap]

        # Spanning-tree inductors
        for i in range(self.num_spn_tree_ind):
            edge_ref    = self.spn_tree_ind_list[i]['ref']

            # Determine dIL by using KCL equation and differentiating.
            # For example, if IL3  =  IL1  + IL2  - Is then
            #                 dIL3 = dIL1 + dIL2 - dIs

            # Extract the KCL equation relating to the co-tree capacitor
            qf_row           = self.ckt.qf[edge_ref,:].copy()

            # Mask references to the co-tree capacitor in the bf row
            qf_row[edge_ref] = 0

            # Multiply by dIL of spanning-tree capacitors
            qf_x             = np.transpose(edge_voltages_x.copy())

            # voltage/current sources (replace with derivatives)
            for i in range(self.num_src_edges):
                src_edge_ref       = self.src_edge_list[i]['ref']
                qf_x[src_edge_ref] = dy_edge_currents_x[src_edge_ref]

            qf_sol = -1 * qf_row.dot(qf_x)
            self.spn_tree_ind_dy[edge_ref] = qf_sol

            edge_voltages_x[edge_ref]    = qf_sol
            edge_currents_x[edge_ref]    = x[edge_ref]

        # spanning-tree resistors
        for i in range(self.num_spn_tree_res):
            edge_ref    = self.spn_tree_res_list[i]['ref']
            edge_info   = self.spn_tree_res_list[i]['info']

            edge_currents_x[edge_ref] = x[edge_ref] / edge_info.get_value(0, 0, 0)

        # co-tree resistors
        for i in range(self.num_co_tree_res):
            edge_ref    = self.co_tree_res_list[i]['ref']
            edge_info   = self.co_tree_res_list[i]['info']

            edge_currents_x[edge_ref] = x[edge_ref] / edge_info.get_value(0, 0, 0)

        # ----------------------------------
        # Qf and Bf
        # ----------------------------------
        # sys var caps
        for i in range(self.num_sys_var_cap):
            edge_ref    = self.sys_cap_edge_list[i]['ref']
            sys_var_ref = self.sys_cap_edge_list[i]['sys_var_ref']
            edge_info   = self.sys_cap_edge_list[i]['info']

            self.cap_value[edge_ref]  = edge_info.get_value(t, edge_voltages_x, edge_currents_x)

            qf_iter[:,edge_ref] = qf_iter[:,edge_ref] * self.cap_value[edge_ref]

        # co-tree caps
        for i in range(self.num_co_tree_cap):
            edge_ref    = self.co_tree_cap_list[i]['ref']
            edge_info   = self.co_tree_cap_list[i]['info']

            self.cap_value[edge_ref]  = edge_info.get_value(t, edge_voltages_x, edge_currents_x)

            qf_iter[:,edge_ref] = qf_iter[:,edge_ref] * self.cap_value[edge_ref]

        # sys var inductors
        for i in range(self.num_sys_var_ind):
            edge_ref    = self.sys_ind_edge_list[i]['ref']
            sys_var_ref = self.sys_ind_edge_list[i]['sys_var_ref']
            edge_info   = self.sys_ind_edge_list[i]['info']

            self.ind_value[edge_ref]  = edge_info.get_value(t, edge_voltages_x, edge_currents_x)

            bf_iter[:,edge_ref] = bf_iter[:,edge_ref] * self.ind_value[edge_ref]

        # spn-tree inductors
        for i in range(self.num_spn_tree_ind):
            edge_ref    = self.spn_tree_ind_list[i]['ref']
            sys_var_ref = self.spn_tree_ind_list[i]['sys_var_ref']
            edge_info   = self.spn_tree_ind_list[i]['info']

            self.ind_value[edge_ref]  = edge_info.get_value(t, edge_voltages_x, edge_currents_x)

            bf_iter[:,edge_ref] = bf_iter[:,edge_ref] * self.ind_value[edge_ref]

        # Evaluate Qf and Bf equations
        qf_res = qf_iter.dot(edge_currents_x)
        bf_res = bf_iter.dot(edge_voltages_x)

        eqn_res = qf_res + bf_res

        # Evaluate dependence constraints
        eqn_res_dep = np.zeros(self.num_dep_srcs)

        for i in range(self.num_dep_srcs):
            edge_ref    = self.dep_src_edge_list[i]['ref']
            dep_var_ref = self.dep_src_edge_list[i]['dep_ref']
            edge_info   = self.dep_src_edge_list[i]['info']

            dep_val     = edge_info.get_value(t, edge_voltages_x, edge_currents_x)
            eqn_res_dep[i] = dep_val - x[dep_var_ref]

        eqn_res = np.concatenate((eqn_res, eqn_res_dep))

        return eqn_res

    def map_1(self, sys_var, eqn_soln, t):

        #########################
        # Map state vector
        #########################
        dy = np.zeros(self.num_sys_var_cap + self.num_sys_var_ind)

        # sys var caps
        for i in range(self.num_sys_var_cap):
            edge_ref    = self.sys_cap_edge_list[i]['ref']
            sys_var_ref = self.sys_cap_edge_list[i]['sys_var_ref']

            dy[sys_var_ref] = eqn_soln[edge_ref]

        # sys var inductors
        for i in range(self.num_sys_var_ind):
            edge_ref    = self.sys_ind_edge_list[i]['ref']
            sys_var_ref = self.sys_ind_edge_list[i]['sys_var_ref']

            dy[sys_var_ref + self.num_sys_var_cap] = eqn_soln[edge_ref]

        #########################
        # Map eqn solutions to edge
        # voltages
        #########################
        edge_voltages_a = eqn_soln.copy()
        edge_currents_a = eqn_soln.copy() * self.xr

        # sys var caps
        for i in range(self.num_sys_var_cap):
            edge_ref    = self.sys_cap_edge_list[i]['ref']
            sys_var_ref = self.sys_cap_edge_list[i]['sys_var_ref']

            edge_voltages_a[edge_ref] = sys_var[sys_var_ref]

            edge_currents_a[edge_ref] = eqn_soln[edge_ref] * self.cap_value[edge_ref]

        # co-tree caps
        for i in range(self.num_co_tree_cap):
            edge_ref    = self.co_tree_cap_list[i]['ref']

            edge_voltages_a[edge_ref] = eqn_soln[edge_ref]

            edge_currents_a[edge_ref] = self.co_tree_cap_dy[edge_ref] * self.cap_value[edge_ref]

        #########################
        # Map eqn solutions to edge
        # currents
        #########################

        # sys var inductors
        for i in range(self.num_sys_var_ind):
            edge_ref    = self.sys_ind_edge_list[i]['ref']
            sys_var_ref = self.sys_ind_edge_list[i]['sys_var_ref']

            edge_currents_a[edge_ref] = sys_var[sys_var_ref + self.num_sys_var_cap]

            edge_voltages_a[edge_ref] = eqn_soln[edge_ref] * self.ind_value[edge_ref]

        # spn-tree inductors
        for i in range(self.num_spn_tree_ind):
            edge_ref    = self.spn_tree_ind_list[i]['ref']

            edge_currents_a[edge_ref] = eqn_soln[edge_ref]

            edge_voltages_a[edge_ref] = self.spn_tree_ind_dy[edge_ref] * self.ind_value[edge_ref]

        #########################
        # Map voltage/current srcs
        #########################

        # voltage/current sources
        for i in range(self.num_src_edges):
            edge_ref    = self.src_edge_list[i]['ref']
            edge_info   = self.src_edge_list[i]['info']

            dependent_edge    = edge_info.get_dependant()

            if edge_info.get_type() == ElementType.voltage_src:
                if dependent_edge:
                    dep_var_ref = self.src_edge_list[i]['dep_ref']
                    edge_voltages_a[edge_ref] = eqn_soln[dep_var_ref]
                else:
                    edge_voltages_a[edge_ref] = edge_info.get_value(0, t, 0)
            else:
                if dependent_edge:
                    dep_var_ref = self.src_edge_list[i]['dep_ref']
                    edge_currents_a[edge_ref] = eqn_soln[dep_var_ref]
                else:
                    edge_currents_a[edge_ref] = edge_info.get_value(0, t, 0)

        edge_currents = np.zeros(self.num_edges)
        edge_voltages = np.zeros(self.num_edges)

        for i in range(self.num_edges):
            edge_currents[i] = edge_currents_a[i]
            edge_voltages[i] = edge_voltages_a[i]

        self.dy            = dy
        self.edge_voltages = edge_voltages
        self.edge_currents = edge_currents

    def dypc_eqn_nonlinear_bsim(self, x, sys_var, t):
        edge_voltages_x     = np.zeros(self.num_edges)
        edge_currents_x     = np.zeros(self.num_edges)

        dy_edge_voltages_x  = np.zeros(self.num_edges)
        dy_edge_currents_x  = np.zeros(self.num_edges)

        self.co_tree_cap_dy  = np.zeros(self.num_edges)
        self.spn_tree_ind_dy = np.zeros(self.num_edges)

        qf_iter = self.ckt.qf.copy()
        bf_iter = self.ckt.bf.copy()

        # ----------------------------------
        # Update edge voltages and currents
        # ----------------------------------
        for i in range(self.num_edges):
            edge_voltages_x[i] = x[i]
            edge_currents_x[i] = x[i]

        # voltage/current sources
        for i in range(self.num_src_edges):
            edge_ref    = self.src_edge_list[i]['ref']
            edge_info   = self.src_edge_list[i]['info']

            dependent_edge    = edge_info.get_dependant()

            if edge_info.get_type() == ElementType.voltage_src:
                if dependent_edge:
                    dep_var_ref = self.src_edge_list[i]['dep_ref']
                    edge_voltages_x[edge_ref]    = x[dep_var_ref]
                    dy_edge_voltages_x[edge_ref] = 0 # REVISIT will need another variable
                else:
                    if edge_ref == 0:
                        edge_voltages_x[edge_ref]    = 0.85 * edge_info.get_value(0, t, 0)
                    else:
                        edge_voltages_x[edge_ref]    = edge_info.get_value(0, t, 0)
                    dy_edge_voltages_x[edge_ref] = edge_info.get_dy_value(0, t, 0)
            #else:
            #    if dependent_edge:
            #        dep_var_ref = self.src_edge_list[i]['dep_ref']
            #        edge_currents_x[edge_ref]    = x[dep_var_ref]
            #        dy_edge_currents_x[edge_ref] = 0 # REVISIT will need another variable
            #    else:
            #        edge_currents_x[edge_ref]    = edge_info.get_value(0, t, 0)
            #        dy_edge_currents_x[edge_ref] = edge_info.get_dy_value(0, t, 0)

        # MOS currents
        delta_t = t - self.t_prev
        if (delta_t == 0):
            delta_t = 1e-12

        edge_currents_x = self.subckt.update_vccs_values(edge_currents_x, 0, delta_t, x)

        # Over-ride for sys var caps
        for i in range(self.num_sys_var_cap):
            edge_ref    = self.sys_cap_edge_list[i]['ref']
            sys_var_ref = self.sys_cap_edge_list[i]['sys_var_ref']

            edge_voltages_x[edge_ref]    = sys_var[sys_var_ref]
            edge_currents_x[edge_ref]    = x[edge_ref]

        # Co-tree capacitors
        for i in range(self.num_co_tree_cap):
            edge_ref    = self.co_tree_cap_list[i]['ref']

            # Determine dVc by using KVL equation and differentiating.
            # For example, if Vc3  =  Vc1  + Vc2  - Vs then
            #                 dVc3 = dVc1 + dVc2 - dVs

            # Extract the KVL equation relating to the co-tree capacitor
            bf_row           = self.ckt.bf[edge_ref,:].copy()

            # Mask references to the co-tree capacitor in the bf row
            bf_row[edge_ref] = 0

            # Multiply by dVc of spanning-tree capacitors
            bf_x             = np.transpose(edge_currents_x.copy())

            # voltage/current sources (replace with derivatives)
            for i in range(self.num_src_edges):
                src_edge_ref       = self.src_edge_list[i]['ref']
                bf_x[src_edge_ref] = dy_edge_voltages_x[src_edge_ref]

            bf_sol = -1 * bf_row.dot(bf_x)
            self.co_tree_cap_dy[edge_ref] = bf_sol

            edge_voltages_x[edge_ref]    = x[edge_ref]
            edge_currents_x[edge_ref]    = bf_sol

        # Over-ride for sys var inductors
        for i in range(self.num_sys_var_ind):
            edge_ref    = self.sys_ind_edge_list[i]['ref']
            sys_var_ref = self.sys_ind_edge_list[i]['sys_var_ref']

            edge_currents_x[edge_ref] = sys_var[sys_var_ref + self.num_sys_var_cap]

        # Spanning-tree inductors
        for i in range(self.num_spn_tree_ind):
            edge_ref    = self.spn_tree_ind_list[i]['ref']

            # Determine dIL by using KCL equation and differentiating.
            # For example, if IL3  =  IL1  + IL2  - Is then
            #                 dIL3 = dIL1 + dIL2 - dIs

            # Extract the KCL equation relating to the co-tree capacitor
            qf_row           = self.ckt.qf[edge_ref,:].copy()

            # Mask references to the co-tree capacitor in the bf row
            qf_row[edge_ref] = 0

            # Multiply by dIL of spanning-tree capacitors
            qf_x             = np.transpose(edge_voltages_x.copy())

            # voltage/current sources (replace with derivatives)
            for i in range(self.num_src_edges):
                src_edge_ref       = self.src_edge_list[i]['ref']
                qf_x[src_edge_ref] = dy_edge_currents_x[src_edge_ref]

            qf_sol = -1 * qf_row.dot(qf_x)
            self.spn_tree_ind_dy[edge_ref] = qf_sol

            edge_voltages_x[edge_ref]    = qf_sol
            edge_currents_x[edge_ref]    = x[edge_ref]

        xr = np.ones(self.num_edges)
        xr = np.transpose(xr)

        # spanning-tree resistors
        for i in range(self.num_spn_tree_res):
            edge_ref    = self.spn_tree_res_list[i]['ref']
            edge_info   = self.spn_tree_res_list[i]['info']

            xr[edge_ref] = 1 / edge_info.get_value(0, 0, 0)

        # co-tree resistors
        for i in range(self.num_co_tree_res):
            edge_ref    = self.co_tree_res_list[i]['ref']
            edge_info   = self.co_tree_res_list[i]['info']

            xr[edge_ref] = 1 / edge_info.get_value(0, 0, 0)

        edge_currents_x = edge_currents_x * xr

        # ----------------------------------
        # Qf and Bf
        # ----------------------------------
        # sys var caps
        for i in range(self.num_sys_var_cap):
            edge_ref    = self.sys_cap_edge_list[i]['ref']
            sys_var_ref = self.sys_cap_edge_list[i]['sys_var_ref']
            edge_info   = self.sys_cap_edge_list[i]['info']

            self.cap_value[edge_ref]  = edge_info.get_value(t, edge_voltages_x, edge_currents_x)

            qf_iter[:,edge_ref] = qf_iter[:,edge_ref] * self.cap_value[edge_ref]

        # co-tree caps
        for i in range(self.num_co_tree_cap):
            edge_ref    = self.co_tree_cap_list[i]['ref']
            edge_info   = self.co_tree_cap_list[i]['info']

            self.cap_value[edge_ref]  = edge_info.get_value(t, edge_voltages_x, edge_currents_x)

            qf_iter[:,edge_ref] = qf_iter[:,edge_ref] * self.cap_value[edge_ref]

        # sys var inductors
        for i in range(self.num_sys_var_ind):
            edge_ref    = self.sys_ind_edge_list[i]['ref']
            sys_var_ref = self.sys_ind_edge_list[i]['sys_var_ref']
            edge_info   = self.sys_ind_edge_list[i]['info']

            self.ind_value[edge_ref]  = edge_info.get_value(t, edge_voltages_x, edge_currents_x)

            bf_iter[:,edge_ref] = bf_iter[:,edge_ref] * self.ind_value[edge_ref]

        # spn-tree inductors
        for i in range(self.num_spn_tree_ind):
            edge_ref    = self.spn_tree_ind_list[i]['ref']
            sys_var_ref = self.spn_tree_ind_list[i]['sys_var_ref']
            edge_info   = self.spn_tree_ind_list[i]['info']

            self.ind_value[edge_ref]  = edge_info.get_value(t, edge_voltages_x, edge_currents_x)

            bf_iter[:,edge_ref] = bf_iter[:,edge_ref] * self.ind_value[edge_ref]

        # Evaluate Qf and Bf equations
        qf_res = qf_iter.dot(edge_currents_x)
        bf_res = bf_iter.dot(edge_voltages_x)

        eqn_res = qf_res + bf_res

        self.bf_res = bf_res
        self.edge_voltages_x = edge_voltages_x

        self.qf_res = qf_res
        self.edge_currents_x = edge_currents_x

        self.in_put = x

        self.qf_iter_m = qf_iter

        # Evaluate dependence constraints
        eqn_res_dep = np.zeros(self.num_dep_srcs)

        for i in range(self.num_dep_srcs):
            edge_ref    = self.dep_src_edge_list[i]['ref']
            dep_var_ref = self.dep_src_edge_list[i]['dep_ref']
            edge_info   = self.dep_src_edge_list[i]['info']

            dep_val     = edge_info.get_value(t, edge_voltages_x, edge_currents_x)
            eqn_res_dep[i] = dep_val - x[dep_var_ref]

        eqn_res = np.concatenate((eqn_res, eqn_res_dep))

        return eqn_res

    def map_bsim(self, sys_var, eqn_soln, t):

        #########################
        # Map state vector
        #########################
        dy = np.zeros(self.num_sys_var_cap + self.num_sys_var_ind)

        # sys var caps
        for i in range(self.num_sys_var_cap):
            edge_ref    = self.sys_cap_edge_list[i]['ref']
            sys_var_ref = self.sys_cap_edge_list[i]['sys_var_ref']

            dy[sys_var_ref] = eqn_soln[edge_ref]

        # sys var inductors
        for i in range(self.num_sys_var_ind):
            edge_ref    = self.sys_ind_edge_list[i]['ref']
            sys_var_ref = self.sys_ind_edge_list[i]['sys_var_ref']

            dy[sys_var_ref + self.num_sys_var_cap] = eqn_soln[edge_ref]

        #########################
        # Map eqn solutions to edge
        # voltages
        #########################
        edge_voltages_a = eqn_soln.copy()
        edge_currents_a = eqn_soln.copy() * self.xr

        # sys var caps
        for i in range(self.num_sys_var_cap):
            edge_ref    = self.sys_cap_edge_list[i]['ref']
            sys_var_ref = self.sys_cap_edge_list[i]['sys_var_ref']

            edge_voltages_a[edge_ref] = sys_var[sys_var_ref]

            edge_currents_a[edge_ref] = eqn_soln[edge_ref] * self.cap_value[edge_ref]

        # co-tree caps
        for i in range(self.num_co_tree_cap):
            edge_ref    = self.co_tree_cap_list[i]['ref']

            edge_voltages_a[edge_ref] = eqn_soln[edge_ref]

            edge_currents_a[edge_ref] = self.co_tree_cap_dy[edge_ref] * self.cap_value[edge_ref]

        #########################
        # Map eqn solutions to edge
        # currents
        #########################

        # sys var inductors
        for i in range(self.num_sys_var_ind):
            edge_ref    = self.sys_ind_edge_list[i]['ref']
            sys_var_ref = self.sys_ind_edge_list[i]['sys_var_ref']

            edge_currents_a[edge_ref] = sys_var[sys_var_ref + self.num_sys_var_cap]

            edge_voltages_a[edge_ref] = eqn_soln[edge_ref] * self.ind_value[edge_ref]

        # spn-tree inductors
        for i in range(self.num_spn_tree_ind):
            edge_ref    = self.spn_tree_ind_list[i]['ref']

            edge_currents_a[edge_ref] = eqn_soln[edge_ref]

            edge_voltages_a[edge_ref] = self.spn_tree_ind_dy[edge_ref] * self.ind_value[edge_ref]

        #########################
        # Map voltage/current srcs
        #########################

        bsim_vccs = np.zeros(self.num_edges)
        self.subckt.get_vccs_values(bsim_vccs)

        # voltage/current sources
        for i in range(self.num_src_edges):
            edge_ref    = self.src_edge_list[i]['ref']
            edge_info   = self.src_edge_list[i]['info']

            dependent_edge    = edge_info.get_dependant()

            if edge_info.get_type() == ElementType.voltage_src:
                if dependent_edge:
                    dep_var_ref = self.src_edge_list[i]['dep_ref']
                    edge_voltages_a[edge_ref] = eqn_soln[dep_var_ref]
                else:
                    edge_voltages_a[edge_ref] = edge_info.get_value(0, t, 0)
            else:
                if dependent_edge:
                    dep_var_ref = self.src_edge_list[i]['dep_ref']
                    edge_currents_a[edge_ref] = eqn_soln[dep_var_ref]
                else:
                    edge_currents_a[edge_ref] = bsim_vccs[edge_ref]

        edge_currents = np.zeros(self.num_edges)
        edge_voltages = np.zeros(self.num_edges)

        for i in range(self.num_edges):
            edge_currents[i] = edge_currents_a[i]
            edge_voltages[i] = edge_voltages_a[i]

        self.dy            = dy
        self.edge_voltages = edge_voltages
        self.edge_currents = edge_currents
