import sys, getopt, re, shutil, os, pprint, networkx as nx

sys.path.append("../tools")
sys.path.append("../subckts")
from roggia import Circuit, ElementType, TwoPortElement
from eqn_utility import EqnUtility
from mos import *
from bsim_cmg import *
import numpy as np
from scipy.integrate import odeint
from scipy.integrate import ode
from scipy.integrate import simps
from scipy import optimize
import matplotlib.pyplot as plt
import cProfile
import struct

import argparse

import process_params
import hp_nmos_16
import hp_pmos_16

def dypc_inverter(t, y, eqn_util):
    tol = 1e-9

    prev_eqn_soln = eqn_util.get_prev_eqn_soln()

    eqn_sol  = optimize.root(eqn_util.dypc_eqn_nonlinear_bsim, prev_eqn_soln, args=(y, t), method='hybr', tol=tol, options={'xtol':tol})
    eqn_soln = eqn_sol.x

    if eqn_sol.success == False:
        print ("Time = " + str(t))
        print (eqn_sol.message)

    eqn_util.map_bsim(y, eqn_soln, t)
    eqn_util.set_prev_eqn_soln(eqn_soln)

    dy_cap = eqn_util.dy
    return dy_cap

def print_info(num_edges, num_sys_vars, num_caps, num_inductors, num_sys_var_cap, num_sys_var_ind):

    print ("-----------------------")
    print ("NUM EDGES       :", num_edges)
    print ("NUM SYS VAR     :", num_sys_vars)
    print ("NUM CAPS        :", num_caps)
    print ("NUM INDS        :", num_inductors)
    print ("NUM SYS VAR CAP :", num_sys_var_cap)
    print ("NUM SYS VAR IND :", num_sys_var_ind)
    print ("-----------------------")

def ode_solve(my_circuit, edge_ref_dict, args, eqn_util):

    num_edges       = my_circuit.get_num_edges()
    num_sys_vars    = my_circuit.get_num_sys_vars()
    num_caps        = my_circuit.get_num_caps()
    num_inductors   = my_circuit.get_num_inductors()
    num_sys_var_cap = my_circuit.get_num_sys_var_cap()
    num_sys_var_ind = my_circuit.get_num_sys_var_ind()

    print_info(num_edges, num_sys_vars, num_caps, num_inductors, num_sys_var_cap, num_sys_var_ind)

    prev_eqn_soln = np.zeros(eqn_util.num_edges + eqn_util.num_dep_srcs)
    eqn_util.set_prev_eqn_soln(prev_eqn_soln)

    edge_dict    = my_circuit.create_edge_dict()

    t0 = 0
    t  = np.linspace(t0, args.tend, args.tstep)
    x0 = np.zeros (num_sys_vars)

    r = ode(dypc_inverter).set_integrator('lsoda', method='bdf', atol=args.abstol, rtol=args.reltol, max_step=args.hmax)
    r.set_initial_value(x0, 0.0)
    r.set_f_params(eqn_util)

    y    = np.empty((int(args.tstep), num_sys_vars))
    y[0] = x0

    i_di_si_p = np.empty((int(args.tstep), num_sys_vars))
    i_si_di_p = np.empty((int(args.tstep), num_sys_vars))
    i_d_di_p  = np.empty((int(args.tstep), num_sys_vars))
    i_di_si_n = np.empty((int(args.tstep), num_sys_vars))
    i_si_di_n = np.empty((int(args.tstep), num_sys_vars))

    k = 1
    while r.successful() and k < args.tstep:
        r.integrate(t[k])

        y[k]       = r.y
        i_di_si_p[k] =  eqn_util.edge_currents[6]
        i_d_di_p[k]  =  eqn_util.edge_currents[8]
        i_si_di_p[k] =  eqn_util.edge_currents[9]
        i_di_si_n[k] =  eqn_util.edge_currents[20]
        i_si_di_n[k] =  eqn_util.edge_currents[22]
        eqn_util.subckt.update_charge()
        eqn_util.set_tprev(t[k])

        k = k + 1

    yz = np.vstack(y).T

    v_in  = np.zeros(len(t))
    v_s   = np.ones(len(t))

    for i in range(len(t)):
        v_in[i]   = hp_nmos_16.vdd * edge_ref_dict['Vin_obj'].get_value(None, t[i], [])

    ax = plt.subplot(3,1,1)
    plt.plot(t, y[:,0], t, v_in)
    #ax.set_ylim([-0.1, 1.1])

    ax = plt.subplot(3,1,2)
    plt.plot(t, i_di_si_n, t, i_si_di_n)
    #ax.set_ylim([-0.1, 1.1])

    plt.subplot(3,1,3)
    plt.plot(t, i_si_di_p, t, i_di_si_p)

    plt.show()

def add_nmos(my_circuit, args):

    Vin_val   = 1.0
    Rin_val   = 0.1e3
    Vs_val    = hp_nmos_16.vdd
    Rs0_val   = 0.1e3
    Cout0_val = 5e-15

    clk_per    = 0.25e-9
    data_slew  = 1e-12

    # Add Vin
    Vin = PulseVoltageSource()
    Vin.set_type(ElementType.voltage_src)
    Vin.set_value(Vin_val)
    Vin.set_instance("Vin")
    Vin.set_pulse_params(data_slew, data_slew, clk_per*4, 0.5, clk_per/2)#clk_per/2
    my_circuit.add_edge(1, 0, Vin)

    # Add Rin
    Rin = TwoPortElement()
    Rin.set_type(ElementType.resistor)
    Rin.set_value(Rin_val)
    Rin.set_instance("Rin")
    my_circuit.add_edge(1, 2, Rin)

    # Add Vs
    Vs = TwoPortElement()
    Vs.set_type(ElementType.voltage_src)
    Vs.set_value(Vs_val)
    Vs.set_instance("Vs")
    my_circuit.add_edge(3, 0, Vs)

    # Add Rs0
    Rs0 = TwoPortElement()
    Rs0.set_type(ElementType.resistor)
    Rs0.set_value(Rs0_val)
    Rs0.set_instance("Rs0")
    my_circuit.add_edge(3, 4, Rs0)

    # Add Cout
    Cout0 = TwoPortElement()
    Cout0.set_type(ElementType.capacitor)
    Cout0.set_value(Cout0_val)
    Cout0.set_instance("Cout0")
    my_circuit.add_edge(5, 0, Cout0)

    # Add inverter 0 (20/1 PMOS 10/1 NMOS)
    inverter0 = InverterBSIM()
    inverter1 = InverterBSIM()
    # PWR, GND, IN, OUT
    default_pmos_params = {'TYPE'        : 0}
    default_nmos_params = {'TYPE'        : 1}

    if args.dflt_prm:
        nmos_params = default_nmos_params
        pmos_params = default_pmos_params
    else:
        nmos_params = hp_nmos_16.params
        pmos_params = hp_pmos_16.params
        pmos_params['L'] = 2 * pmos_params['L']

    inverter0.add_edges(my_circuit, 4, 0, 2, 5, nmos_params, pmos_params, 1)


    dict = {'Vin_ref'   : 0, 'Vin_val'   : Vin_val,   'Vin_obj'  : Vin,  \
            'Rin_ref'   : 1, 'Rin_val'   : Rin_val,    \
            'Vs_ref'    : 2, 'Vs_val'    : Vs_val,    'Vs_obj'   : Vs, \
            'Rs0_ref'   : 3, 'Rs0_val'   : Rs0_val,    \
            'Cout0_ref' : 4, 'Cout0_val' : Cout0_val,  \
            'inverter0_obj': inverter0}

    return dict

def parseArguments():

    parser = argparse.ArgumentParser()

    parser.add_argument("-reltol",   help="Relative Tolerance",   type=float, default=1e-3)
    parser.add_argument("-abstol",   help="Absolute Tolerance",   type=float, default=1e-3)
    parser.add_argument("-hmax",     help="Max Step Size",        type=float, default=0.01e-9)
    parser.add_argument("-tend",     help="Simulation End Time",  type=float, default=4e-9)
    parser.add_argument("-tstep",    help="Simulation Step Time", type=float, default=10e3)
    parser.add_argument("-level",    help="MOSFET model",         type=int,   default=1)
    parser.add_argument("-profile",  help="Run with cProfile",    action='store_true')
    parser.add_argument("-dflt_prm", help="Default MOS params",   action='store_true')

    args = parser.parse_args()

    return args

def main(args):

    test_circuit = Circuit()
    edge_ref_dict = add_nmos(test_circuit, args)

    test_circuit.init_circuit()

    eqn_util = EqnUtility(test_circuit)
    eqn_util.add_subckt(edge_ref_dict['inverter0_obj'])

    global edge_dict
    edge_dict = test_circuit.create_edge_dict()

    if args.profile:
        cProfile.runctx('ode_solve(test_circuit, edge_ref_dict, args)', {'test_circuit': test_circuit,   \
                                                                         'edge_ref_dict': edge_ref_dict, \
                                                                         'args': args,                   \
                                                                         'ode_solve': ode_solve}, {})
    else:
        ode_solve(test_circuit, edge_ref_dict, args, eqn_util)

if __name__ == "__main__":

    args = parseArguments()
    main(args)
