from roggia import Circuit, ElementType, TwoPortElement

import sys
sys.path.append("../sources")
sys.path.append("../models")
import sources
import models
import numpy as np
import math
from bsim_cmg import *

class variac(TwoPortElement):

    def __init__(self):
        self.type     = ElementType.capacitor

    def get_value(self, t, voltages):
        value = 2 * self.value * voltages[self.ctrl_vref] + (self.value / 4)

        return value

    def set_value(self, value):
        self.value     = value

    def set_ctrl_vref(self, ctrl_vref):
        self.ctrl_vref = ctrl_vref

class variac_isrc(TwoPortElement):

    def __init__(self):
        self.type           = ElementType.current_src
        self.current_charge = 0.0
        self.last_charge    = 0.0
        self.t0             = 0.0

        self.dependant      = False

    def get_value_1(self, t, voltages, currents):
        value = self.value + self.value * voltages[self.ctrl_vref]**2 + (self.value * 3.15) * voltages[self.ctrl_vref]

        return value

    def get_value(self, t, voltages, currents):
        value  = self.value * voltages[self.ctrl_vref]**2  + ((self.value / 4) * voltages[self.ctrl_vref])

        dy = value - self.last_charge

        self.current_charge = value
        self.current_t      = t

        if t == self.t0:
            dt = 1e-9
        else:
            dt = t - self.t0

        return (dy / dt)

    def update_charge(self):
        self.last_charge = self.current_charge
        self.t0          = self.current_t

    def set_initial_charge(self, voltages):
        self.get_value(0, voltages, 0)
        self.last_charge = self.current_charge
        self.t0          = 0

    def set_value(self, value):
        self.value     = value

    def set_ctrl_vref(self, ctrl_vref):
        self.ctrl_vref = ctrl_vref

class MosDigitalCIN(TwoPortElement):

    def __init__(self):
        self.type     = ElementType.capacitor
        self.L        = 0.0
        self.W        = 0.0
        self.scale    = 0.0
        self.cox_dash = 0.0
        self.dynamic  = False

    def set_value(self, L, W, scale, cox_dash):
        Cox        = cox_dash * (scale**2 / (1e-6)**2) * W * L
        self.value = (3/2) * Cox

    def get_static_value(self):
        return self.value

class MosDigitalCOUT(TwoPortElement):

    def __init__(self):
        self.type     = ElementType.capacitor
        self.L        = 0.0
        self.W        = 0.0
        self.scale    = 0.0
        self.cox_dash = 0.0
        self.dynamic  = False

    def set_value(self, L, W, scale, cox_dash):
        Cox        = cox_dash * (scale**2 / (1e-6)**2) * W * L
        self.value = Cox

    def get_static_value(self):
        return self.value

class NmosDigitalIDS(TwoPortElement):

    def __init__(self):
        self.type      = ElementType.current_src
        self.L         = 0.0
        self.W         = 0.0
        self.scale     = 0.0
        self.cox_dash  = 0.0
        self.cin_ref   = 0
        self.cout_ref  = 0
        self.dynamic   = True
        self.dependant = True
        self.MAXP      =  1e-3
        self.MAXN      = -1e-3
        self.KPn       = 120e-6
        self.Vthn      = 280e-3

    def set_value(self, L, W, scale, cox_dash, cin_ref, cout_ref):
        self.L        = L
        self.W        = W
        self.scale    = scale
        self.cox_dash = cox_dash
        self.cin_ref  = cin_ref
        self.cout_ref = cout_ref
        self.MAXP     =       (scale * (self.W / self.L)) * ((1 - 0.5) ** 2)
        self.MAXN     = -1 * ((scale * (self.W / self.L)) * ((1 - 0.5) ** 2))
        self.MAXP     =  700e-6
        self.MAXN     = -700e-6
        self.KPn      = 120e-6
        self.Vthn     = 280e-3

    def get_value(self, host_circuit, t, voltages):

        Vgs = voltages[self.cin_ref]
        Vds = voltages[self.cout_ref]

        # Ohmic
        if Vgs >= self.Vthn and Vds <= (Vgs - self.Vthn):
            ids = self.KPn * (self.W / self.L) * ( (Vgs - self.Vthn) * Vds - (Vds**2 / 2) )
        # Saturation
        elif Vds >= (Vgs - self.Vthn) and Vgs >= self.Vthn:
            ids = (self.KPn / 2) * (self.W / self.L) * (Vgs - self.Vthn)**2
        else:
            ids = 0

        return ids

class PmosDigitalIDS(TwoPortElement):

    def __init__(self):
        self.type      = ElementType.current_src
        self.L         = 0.0
        self.W         = 0.0
        self.scale     = 0.0
        self.cox_dash  = 0.0
        self.cin_ref   = 0
        self.cout_ref  = 0
        self.dynamic   = True
        self.dependant = True
        self.MAXP      =  1e-3
        self.MAXN      = -1e-3
        self.KPp       = 40e-6
        self.Vthp      = 280e-3

    def set_value(self, L, W, scale, cox_dash, cin_ref, cout_ref):
        self.L        = L
        self.W        = W
        self.scale    = scale
        self.cox_dash = cox_dash
        self.cin_ref  = cin_ref
        self.cout_ref = cout_ref
        self.MAXP     =       (scale * (self.W / self.L)) * ((1 - 0.5) ** 2)
        self.MAXN     = -1 * ((scale * (self.W / self.L)) * ((1 - 0.5) ** 2))
        self.MAXP     =  700e-6
        self.MAXN     = -700e-6
        self.KPp      = 40e-6
        self.Vthp     = 280e-3

    def get_value(self, host_circuit, t, voltages):

        Vsg = voltages[self.cout_ref] - voltages[self.cin_ref]
        Vsd = voltages[self.cout_ref]

        # Ohmic
        if Vsg >= self.Vthp and Vsd <= (Vsg - self.Vthp):
            isd = self.KPp * (self.W / self.L) * ( (Vsg - self.Vthp) * Vsd - (Vsd**2 / 2) )
        # Saturation
        elif Vsd >= (Vsg - self.Vthp) and Vsg >= self.Vthp:
            isd = (self.KPp / 2) * (self.W / self.L) * (Vsg - self.Vthp)**2
        else:
            isd = 0

        return isd

class PulseVoltageSource(TwoPortElement):
    def __init__(self):
        self.type      = ElementType.voltage_src
        self.dynamic   = True
        self.dependant = False

    def get_value(self, host_circuit, t, voltages):
        return sources.pulse(self.per, self.t_on, self.rise, self.fall, t + self.offset)

    def get_dy_value(self, host_circuit, t, voltages):
        h   = 1e-9
        t_a = t - h

        v_a = sources.pulse(self.per, self.t_on, self.rise, self.fall, t_a + self.offset)
        v_b = sources.pulse(self.per, self.t_on, self.rise, self.fall, t   + self.offset)

        return (v_b - v_a) / h

    def get_edges(self, t_end):
        return sources.pulse_edge(self.per, self.t_on, self.offset, t_end)

    def set_pulse_params(self, rise, fall, per, duty, offset):
        self.rise   = rise
        self.fall   = fall
        self.per    = per
        self.t_on   = per * duty
        self.offset = offset

class SineVoltageSource(TwoPortElement):
    def __init__(self):
        self.type      = ElementType.voltage_src
        self.dynamic   = True
        self.dependant = False

    def get_value(self, host_circuit, t, voltages):

        return np.sin(2*np.pi*1e6*t)

    def get_dy_value(self, host_circuit, t, voltages):

        return 2*np.pi*1e6 * np.cos(2*np.pi*1e6*t)

###################
# NMOS Level 0
###################
class NmosDigital():

    def __init__(self):
        self.cin_ref  = 0
        self.cout_ref = 0
        self.ids_ref  = 0

    def add_edges(self, ckt, L, W, scale, cox_dash, gate, source, drain):

        # Get the start edge reference from the target Circuit
        ser = ckt.get_num_edges()

        # Add edges in order IDS, CIN, COUT so that edge references are:
        # IDS  = ser
        # CIN  = ser + 1
        # COUT = ser + 2
        self.ids_ref  = ser
        self.cin_ref  = ser + 1
        self.cout_ref = ser + 2

        # Add IDS
        self.ids = NmosDigitalIDS()
        self.ids.set_value(L, W, scale, cox_dash, self.cin_ref, self.cout_ref)
        self.ids.set_instance("nmos_ids")
        ckt.add_edge(drain, source, self.ids)

        # Add CIN
        self.cin = MosDigitalCIN()
        self.cin.set_value(L, W, scale, cox_dash)
        self.cin.set_instance("nmos_cin")
        ckt.add_edge(gate, source, self.cin)

        # Add COUT
        self.cout = MosDigitalCOUT()
        self.cout.set_value(L, W, scale, cox_dash)
        self.cout.set_instance("nmos_cout")
        ckt.add_edge(drain, source, self.cout)

    def update_cap_values(self, cap_values, edge_dict):

        # CIN
        val = self.cin.get_value(None, 0, [])
        cap_values[edge_dict[self.cin_ref]['cap_ref']] = val

        # COUT
        val = self.cout.get_value(None, 0, [])
        cap_values[edge_dict[self.cout_ref]['cap_ref']] = val

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):

        # IDS
        val = self.ids.get_value(None, t, edge_voltages)
        vccs_values[self.ids_ref] = val

        return vccs_values

###################
# PMOS Level 0
###################
class PmosDigital():

    def __init__(self):
        self.cin_ref  = 0
        self.cout_ref = 0
        self.isd_ref  = 0

    def add_edges(self, ckt, L, W, scale, cox_dash, gate, source, drain):

        # Get the start edge reference from the target Circuit
        ser = ckt.get_num_edges()

        # Add edges in order IDS, CIN, COUT so that edge references are:
        # ISD  = ser
        # CIN  = ser + 1
        # COUT = ser + 2
        self.isd_ref  = ser
        self.cin_ref  = ser + 1
        self.cout_ref = ser + 2

        # Add IDS
        self.isd = PmosDigitalIDS()
        self.isd.set_value(L, W, scale, cox_dash, self.cin_ref, self.cout_ref)
        self.isd.set_instance("pmos_isd")
        ckt.add_edge(source, drain, self.isd)

        # Add CIN
        self.cin = MosDigitalCIN()
        self.cin.set_value(L, W, scale, cox_dash)
        self.cin.set_instance("pmos_cin")
        ckt.add_edge(gate, drain, self.cin)

        # Add COUT
        self.cout = MosDigitalCOUT()
        self.cout.set_value(L, W, scale, cox_dash)
        self.cout.set_instance("pmos_cout")
        ckt.add_edge(source, drain, self.cout)

    def update_cap_values(self, cap_values, edge_dict):
        # CIN
        val = self.cin.get_value(None, 0, [])
        cap_values[edge_dict[self.cin_ref]['cap_ref']] = val

        # COUT
        val = self.cout.get_value(None, 0, [])
        cap_values[edge_dict[self.cout_ref]['cap_ref']] = val

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):

        # ISD
        val = self.isd.get_value(None, t, edge_voltages)
        vccs_values[self.isd_ref] = val

        return vccs_values

###################
# NMOS Level 1
###################
class NmosDigitalL1():

    def __init__(self):
        self.cgd_ref  = 0
        self.cdb_ref  = 0
        self.cgb_ref  = 0
        self.cgs_ref  = 0
        self.csb_ref  = 0
        self.ids_ref  = 0

    def add_edges(self, ckt, L, W, scale, cox_dash, params, gate, source, drain, bulk):

        self.L        = L
        self.W        = W
        self.scale    = scale
        self.cox_dash = cox_dash
        self.KPn      = 120e-6
        self.Vthn     = 280e-3
        self.params   = params
        self.Leff     = L * scale

        # Get the start edge reference from the target Circuit
        ser = ckt.get_num_edges()

        # Add edges
        self.ids_ref  = ser
        self.cgd_ref  = ser + 1
        self.cdb_ref  = ser + 2
        self.cgb_ref  = ser + 3
        self.cgs_ref  = ser + 4
        self.csb_ref  = ser + 5

        # Add IDS
        self.ids = TwoPortElement()
        self.ids.set_instance("nmos_ids")
        self.ids.set_type(ElementType.current_src)
        ckt.add_edge(drain, source, self.ids)

        # Add Cgd
        self.cgd = TwoPortElement()
        self.cgd.set_instance("nmos_cgd")
        self.cgd.set_type(ElementType.capacitor)
        ckt.add_edge(gate, drain, self.cgd)

        # Add Cdb
        self.cdb = TwoPortElement()
        self.cdb.set_instance("nmos_cdb")
        self.cdb.set_type(ElementType.capacitor)
        ckt.add_edge(drain, bulk, self.cdb)

        # Add Cgb
        self.cgb = TwoPortElement()
        self.cgb.set_instance("nmos_cgb")
        self.cgb.set_type(ElementType.capacitor)
        ckt.add_edge(gate, bulk, self.cgb)

        # Add Cgs
        self.cgs = TwoPortElement()
        self.cgs.set_instance("nmos_cgs")
        self.cgs.set_type(ElementType.capacitor)
        ckt.add_edge(gate, source, self.cgs)

        # Add Csb
        self.csb = TwoPortElement()
        self.csb.set_instance("nmos_csb")
        self.csb.set_type(ElementType.capacitor)
        ckt.add_edge(source, bulk, self.csb)

    def update_cap_values(self, cap_values, edge_dict, edge_voltages):

        Vgs = edge_voltages[self.cgs_ref]
        Vds = edge_voltages[self.cgs_ref] - edge_voltages[self.cgd_ref]
        Vsb = edge_voltages[self.csb_ref]
        Vdb = edge_voltages[self.cdb_ref]

        #Cjd = (self.params['CJ'] * self.params['AD'] * self.scale**2) / \
        #      (1 + (Vdb / self.params['PB']))**self.params['MJ'] + \
        #      (self.params['CJSW'] * self.params['PD'] * self.scale**2) / \
        #      (1 + (Vdb / self.params['PBSW']))**self.params['MJSW']
        #
        #Cjs = (self.params['CJ'] * self.params['AS'] * self.scale**2) / \
        #      (1 + (Vsb / self.params['PB']))**self.params['MJ'] + \
        #      (self.params['CJSW'] * self.params['PS'] * self.scale**2) / \
        #      (1 + (Vsb / self.params['PBSW']))**self.params['MJSW']

        W    = self.W * self.scale
        L    = self.L * self.scale
        Leff = (self.L - 2 * self.params['LD']) * self.scale

        Cjd = self.params['CJD'] * W * L
        Cjs = self.params['CJS'] * W * L

        if (Vsb > 0):
            Vthn = self.params['VTO'] + self.params['GAMMA'] * (math.sqrt(self.params['PHI'] + Vsb) - math.sqrt(self.params['PHI']))
        else:
            Vthn = self.params['VTO']

        # Ohmic (Triode)
        if Vgs >= Vthn and Vds <= (Vgs - Vthn):

            # Cgd
            val = (0.5 * W * L * self.cox_dash) + self.params['CGDO'] * W
            cap_values[edge_dict[self.cgd_ref]['cap_ref']] = val

            # Cdb
            val = Cjd
            cap_values[edge_dict[self.cdb_ref]['cap_ref']] = val

            # Cgb
            val = self.params['CGBO'] * L
            cap_values[edge_dict[self.cgb_ref]['cap_ref']] = val

            # Cgs
            val = (0.5 * W * L * self.cox_dash) + self.params['CGSO'] * W
            cap_values[edge_dict[self.cgs_ref]['cap_ref']] = val

            # Csb
            val = Cjs
            cap_values[edge_dict[self.csb_ref]['cap_ref']] = val

        # Saturation
        elif Vds >= (Vgs - Vthn) and Vgs >= Vthn:

            # Cgd
            val = self.params['CGDO'] * W
            cap_values[edge_dict[self.cgd_ref]['cap_ref']] = val

            # Cdb
            val = Cjd
            cap_values[edge_dict[self.cdb_ref]['cap_ref']] = val

            # Cgb
            val = self.params['CGBO'] * L
            cap_values[edge_dict[self.cgb_ref]['cap_ref']] = val

            # Cgs
            val = ((2/3) * W * L * self.cox_dash) + self.params['CGSO'] * W
            cap_values[edge_dict[self.cgs_ref]['cap_ref']] = val

            # Csb
            val = Cjs
            cap_values[edge_dict[self.csb_ref]['cap_ref']] = val

        else:

            # Cgd
            val = self.params['CGDO'] * W
            cap_values[edge_dict[self.cgd_ref]['cap_ref']] = val

            # Cdb
            val = Cjd
            cap_values[edge_dict[self.cdb_ref]['cap_ref']] = val

            # Cgb
            val = self.cox_dash * W * Leff + self.params['CGBO'] * L
            cap_values[edge_dict[self.cgb_ref]['cap_ref']] = val

            # Cgs
            val = self.params['CGSO'] * W
            cap_values[edge_dict[self.cgs_ref]['cap_ref']] = val

            # Csb
            val = Cjs
            cap_values[edge_dict[self.csb_ref]['cap_ref']] = val

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):

        Vgs = edge_voltages[self.cgs_ref]
        Vds = edge_voltages[self.cgs_ref] - edge_voltages[self.cgd_ref]
        Vsb = edge_voltages[self.csb_ref]

        if (Vsb > 0):
            Vthn = self.params['VTO'] + self.params['GAMMA'] * (math.sqrt(self.params['PHI'] + Vsb) - math.sqrt(self.params['PHI']))
        else:
            Vthn = self.params['VTO']

        Vds_sat = Vgs - Vthn

        ID0 = 7.1e-9
        q   = 1.6e-19
        K   = 1.38e-23
        T   = 300
        n   = 1

        # Ohmic (Triode)
        if Vgs >= Vthn and Vds <= (Vgs - Vthn):
            ids = self.KPn * (self.W / self.L) * ( (Vgs - Vthn) * Vds - (Vds**2 / 2) )
        # Saturation
        elif Vds >= Vds_sat and Vgs >= Vthn:
            ids = (self.KPn / 2) * (self.W / self.L) * ((Vgs - Vthn)**2) * (1 + self.params['LAMBDA'] * (Vds - Vds_sat))
        else:
            ids = ID0 * (self.W / self.L) * math.exp((q * (Vgs - Vthn)) / (n * K * T))

        vccs_values[self.ids_ref] = ids

        self.ids = ids

        return vccs_values

    def get_vccs_values(self, vccs_values):
        vccs_values[self.ids_ref] = self.ids

        return vccs_values

    def update_vccs_mask(self, vccs_mask):

        vccs_mask[:,self.ids_ref] = 0

        return vccs_mask

    def is_dy_dep(self, edge_ref):
        return (self.cgs_ref == edge_ref) or \
               (self.cgd_ref == edge_ref) or \
               (self.csb_ref == edge_ref)

###################
# PMOS Level 1
###################
class PmosDigitalL1():

    def __init__(self):
        self.cgd_ref  = 0
        self.cdb_ref  = 0
        self.cgb_ref  = 0
        self.cgs_ref  = 0
        self.csb_ref  = 0
        self.isd_ref  = 0

    def add_edges(self, ckt, L, W, scale, cox_dash, params, gate, source, drain, bulk):

        self.L         = L
        self.W         = W
        self.scale     = scale
        self.cox_dash  = cox_dash
        self.KPp       = 40e-6
        self.Vthp      = 280e-3
        self.params    = params
        self.Leff      = L * scale

        # Get the start edge reference from the target Circuit
        ser = ckt.get_num_edges()

        # Add edges
        self.isd_ref  = ser
        self.cgd_ref  = ser + 1
        self.cdb_ref  = ser + 2
        self.cgb_ref  = ser + 3
        self.cgs_ref  = ser + 4
        self.csb_ref  = ser + 5

        # Add ISD
        self.isd = TwoPortElement()
        self.isd.set_instance("nmos_ids")
        self.isd.set_type(ElementType.current_src)
        ckt.add_edge(source, drain, self.isd)

        # Add Cgd
        self.cgd = TwoPortElement()
        self.cgd.set_instance("nmos_cgd")
        self.cgd.set_type(ElementType.capacitor)
        ckt.add_edge(gate, drain, self.cgd)

        # Add Cdb
        self.cdb = TwoPortElement()
        self.cdb.set_instance("nmos_cdb")
        self.cdb.set_type(ElementType.capacitor)
        ckt.add_edge(drain, bulk, self.cdb)

        # Add Cgb
        self.cgb = TwoPortElement()
        self.cgb.set_instance("nmos_cgb")
        self.cgb.set_type(ElementType.capacitor)
        ckt.add_edge(gate, bulk, self.cgb)

        # Add Cgs
        self.cgs = TwoPortElement()
        self.cgs.set_instance("nmos_cgs")
        self.cgs.set_type(ElementType.capacitor)
        ckt.add_edge(gate, source, self.cgs)

        # Add Csb
        self.csb = TwoPortElement()
        self.csb.set_instance("nmos_csb")
        self.csb.set_type(ElementType.capacitor)
        ckt.add_edge(source, bulk, self.csb)

    def update_cap_values(self, cap_values, edge_dict, edge_voltages):

        Vsg = -1 * edge_voltages[self.cgs_ref]
        Vsd = edge_voltages[self.cgd_ref] - edge_voltages[self.cgs_ref]
        Vbs = -1 * edge_voltages[self.csb_ref]
        Vbd = -1 * edge_voltages[self.cdb_ref]

        #Cjd = (self.params['CJ'] * self.params['AD'] * self.scale**2) / \
        #      (1 + (Vbd / self.params['PB']))**self.params['MJ'] + \
        #      (self.params['CJSW'] * self.params['PD'] * self.scale**2) / \
        #      (1 + (Vbd / self.params['PBSW']))**self.params['MJSW']
        #
        #Cjs = (self.params['CJ'] * self.params['AS'] * self.scale**2) / \
        #      (1 + (Vbs / self.params['PB']))**self.params['MJ'] + \
        #      (self.params['CJSW'] * self.params['PS'] * self.scale**2) / \
        #      (1 + (Vbs / self.params['PBSW']))**self.params['MJSW']

        W    = self.W * self.scale
        L    = self.L * self.scale
        Leff = (self.L - 2 * self.params['LD']) * self.scale

        Cjd = self.params['CJD'] * W * L
        Cjs = self.params['CJS'] * W * L

        if (Vbs > 0):
            Vthp = self.params['VTO'] - self.params['GAMMA'] * (math.sqrt(self.params['PHI'] + Vbs) - math.sqrt(self.params['PHI']))
        else:
            Vthp = self.params['VTO']

        Vthp = abs(Vthp)

        # Ohmic (Triode)
        if Vsg >= Vthp and Vsd <= (Vsg - Vthp):

            # Cgd
            val = (0.5 * W * L * self.cox_dash) + self.params['CGDO'] * W
            cap_values[edge_dict[self.cgd_ref]['cap_ref']] = val

            # Cdb
            val = Cjd
            cap_values[edge_dict[self.cdb_ref]['cap_ref']] = val

            # Cgb
            val = self.params['CGBO'] * L
            cap_values[edge_dict[self.cgb_ref]['cap_ref']] = val

            # Cgs
            val = (0.5 * W * L * self.cox_dash) + self.params['CGSO'] * W
            cap_values[edge_dict[self.cgs_ref]['cap_ref']] = val

            # Csb
            val = Cjs
            cap_values[edge_dict[self.csb_ref]['cap_ref']] = val

        # Saturation
        elif Vsd >= (Vsg - Vthp) and Vsg >= Vthp:

            # Cgd
            val = self.params['CGDO'] * W
            cap_values[edge_dict[self.cgd_ref]['cap_ref']] = val

            # Cdb
            val = Cjd
            cap_values[edge_dict[self.cdb_ref]['cap_ref']] = val

            # Cgb
            val = self.params['CGBO'] * L
            cap_values[edge_dict[self.cgb_ref]['cap_ref']] = val

            # Cgs
            val = ((2/3) * W * L * self.cox_dash) + self.params['CGSO'] * W
            cap_values[edge_dict[self.cgs_ref]['cap_ref']] = val

            # Csb
            val = Cjs
            cap_values[edge_dict[self.csb_ref]['cap_ref']] = val

        else:

            # Cgd
            val = self.params['CGDO'] * W
            cap_values[edge_dict[self.cgd_ref]['cap_ref']] = val

            # Cdb
            val = Cjd
            cap_values[edge_dict[self.cdb_ref]['cap_ref']] = val

            # Cgb
            val = self.cox_dash * W * Leff + self.params['CGBO'] * L
            cap_values[edge_dict[self.cgb_ref]['cap_ref']] = val

            # Cgs
            val = self.params['CGSO'] * W
            cap_values[edge_dict[self.cgs_ref]['cap_ref']] = val

            # Csb
            val = Cjs
            cap_values[edge_dict[self.csb_ref]['cap_ref']] = val

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):

        Vsg = -1 * edge_voltages[self.cgs_ref]
        Vsd = edge_voltages[self.cgd_ref] - edge_voltages[self.cgs_ref]
        Vbs = -1 * edge_voltages[self.csb_ref]

        if (Vbs > 0):
            Vthp = self.params['VTO'] - self.params['GAMMA'] * (math.sqrt(self.params['PHI'] + Vbs) - math.sqrt(self.params['PHI']))
        else:
            Vthp = self.params['VTO']

        Vthp = abs(Vthp)
        Vsd_sat = Vsg - Vthp

        ID0 = 10e-9
        q   = 1.6e-19
        K   = 1.38e-23
        T   = 300
        n   = 1

        # Ohmic
        if Vsg >= Vthp and Vsd <= (Vsg - Vthp):
            isd = self.KPp * (self.W / self.L) * ( (Vsg - Vthp) * Vsd - (Vsd**2 / 2) )
        # Saturation
        elif Vsd >= Vsd_sat and Vsg >= Vthp:
            isd = (self.KPp / 2) * (self.W / self.L) * ((Vsg - Vthp)**2) * (1 + self.params['LAMBDA'] * (Vsd - Vsd_sat))
        else:
            isd = ID0 * (self.W / self.L) * math.exp((q * (Vsg - Vthp)) / (n * K * T))

        vccs_values[self.isd_ref] = isd

        self.isd = isd

        return vccs_values

    def get_vccs_values(self, vccs_values):
        vccs_values[self.isd_ref] = self.isd

        return vccs_values

    def update_vccs_mask(self, vccs_mask):

        vccs_mask[:,self.isd_ref] = 0

        return vccs_mask

    def is_dy_dep(self, edge_ref):
        return (self.cgs_ref == edge_ref) or \
               (self.cgd_ref == edge_ref) or \
               (self.csb_ref == edge_ref)

###################
# NAND
###################
class NandDigital():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, Ln, Wn, Lp, Wp, scale, cox_dash, pwr, gnd, in_a, in_b, out_b):

        int_1 = ckt.get_internal_node()

        self.pmos_1 = PmosDigital()
        self.pmos_1.add_edges(ckt, Lp, Wp, scale, cox_dash, in_a, pwr, out_b)

        self.pmos_2 = PmosDigital()
        self.pmos_2.add_edges(ckt, Lp, Wp, scale, cox_dash, in_b, pwr, out_b)

        self.nmos_1 = NmosDigital()
        self.nmos_1.add_edges(ckt, Ln, Wn, scale, cox_dash, in_a, int_1, out_b)

        self.nmos_2 = NmosDigital()
        self.nmos_2.add_edges(ckt, Ln, Wn, scale, cox_dash, in_b, gnd, int_1)

    def update_cap_values(self, cap_values, edge_dict):
        cap_values = self.pmos_1.update_cap_values(cap_values, edge_dict)
        cap_values = self.pmos_2.update_cap_values(cap_values, edge_dict)
        cap_values = self.nmos_1.update_cap_values(cap_values, edge_dict)
        cap_values = self.nmos_2.update_cap_values(cap_values, edge_dict)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        vccs_values = self.pmos_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values

    def get_vccs_values_ref(self):

        vccs_values_ref = []

        vccs_values_ref.append(self.pmos_1.update_vccs_values)
        vccs_values_ref.append(self.pmos_2.update_vccs_values)
        vccs_values_ref.append(self.nmos_1.update_vccs_values)
        vccs_values_ref.append(self.nmos_2.update_vccs_values)

        return vccs_values_ref

###################
# NAND L1
###################
class NandDigitalL1():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, in_a, in_b, out_b):

        int_1 = ckt.get_internal_node()

        self.pmos_1 = PmosDigitalL1()
        self.pmos_1.add_edges(ckt, Lp, Wp, scale, cox_dash, p_params, in_a, pwr, out_b, pwr)

        self.pmos_2 = PmosDigitalL1()
        self.pmos_2.add_edges(ckt, Lp, Wp, scale, cox_dash, p_params, in_b, pwr, out_b, pwr)

        self.nmos_1 = NmosDigitalL1()
        self.nmos_1.add_edges(ckt, Ln, Wn, scale, cox_dash, n_params, in_a, int_1, out_b, gnd)

        self.nmos_2 = NmosDigitalL1()
        self.nmos_2.add_edges(ckt, Ln, Wn, scale, cox_dash, n_params, in_b, gnd, int_1, gnd)

    def update_cap_values(self, cap_values, edge_dict, edge_voltages):
        cap_values = self.pmos_1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.pmos_2.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_2.update_cap_values(cap_values, edge_dict, edge_voltages)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        vccs_values = self.pmos_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values

###################
# AND
###################
class AndDigital():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, Ln, Wn, Lp, Wp, scale, cox_dash, pwr, gnd, in_a, in_b, out_b):

        int_1 = ckt.get_internal_node()
        int_2 = ckt.get_internal_node()

        # NAND
        self.pmos_1 = PmosDigital()
        self.pmos_1.add_edges(ckt, Lp, Wp, scale, cox_dash, in_a, pwr, int_2)

        self.pmos_2 = PmosDigital()
        self.pmos_2.add_edges(ckt, Lp, Wp, scale, cox_dash, in_b, pwr, int_2)

        self.nmos_1 = NmosDigital()
        self.nmos_1.add_edges(ckt, Ln, Wn, scale, cox_dash, in_a, int_1, int_2)

        self.nmos_2 = NmosDigital()
        self.nmos_2.add_edges(ckt, Ln, Wn, scale, cox_dash, in_b, gnd, int_1)

        # Inverter
        self.pmos_3 = PmosDigital()
        self.pmos_3.add_edges(ckt, Lp, Wp, scale, cox_dash, int_2, pwr, out_b)

        self.nmos_3 = NmosDigital()
        self.nmos_3.add_edges(ckt, Ln, Wn, scale, cox_dash, int_2, gnd, out_b)

    def update_cap_values(self, cap_values, edge_dict):
        cap_values = self.pmos_1.update_cap_values(cap_values, edge_dict)
        cap_values = self.pmos_2.update_cap_values(cap_values, edge_dict)
        cap_values = self.pmos_3.update_cap_values(cap_values, edge_dict)
        cap_values = self.nmos_1.update_cap_values(cap_values, edge_dict)
        cap_values = self.nmos_2.update_cap_values(cap_values, edge_dict)
        cap_values = self.nmos_3.update_cap_values(cap_values, edge_dict)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        vccs_values = self.pmos_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_3.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_3.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values

    def get_vccs_values_ref(self):

        vccs_values_ref = []

        vccs_values_ref.append(self.pmos_1.update_vccs_values)
        vccs_values_ref.append(self.pmos_2.update_vccs_values)
        vccs_values_ref.append(self.pmos_3.update_vccs_values)
        vccs_values_ref.append(self.nmos_1.update_vccs_values)
        vccs_values_ref.append(self.nmos_2.update_vccs_values)
        vccs_values_ref.append(self.nmos_3.update_vccs_values)

        return vccs_values_ref

###################
# AND L1
###################
class AndDigitalL1():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, in_a, in_b, out_b):

        int_1 = ckt.get_internal_node()
        int_2 = ckt.get_internal_node()

        # NAND
        self.pmos_1 = PmosDigitalL1()
        self.pmos_1.add_edges(ckt, Lp, Wp, scale, cox_dash, p_params, in_a, pwr, int_2, pwr)

        self.pmos_2 = PmosDigitalL1()
        self.pmos_2.add_edges(ckt, Lp, Wp, scale, cox_dash, p_params, in_b, pwr, int_2, pwr)

        self.nmos_1 = NmosDigitalL1()
        self.nmos_1.add_edges(ckt, Ln, Wn, scale, cox_dash, n_params, in_a, int_1, int_2, gnd)

        self.nmos_2 = NmosDigitalL1()
        self.nmos_2.add_edges(ckt, Ln, Wn, scale, cox_dash, n_params, in_b, gnd, int_1, gnd)

        # Inverter
        self.pmos_3 = PmosDigitalL1()
        self.pmos_3.add_edges(ckt, Lp, Wp, scale, cox_dash, p_params, int_2, pwr, out_b, pwr)

        self.nmos_3 = NmosDigitalL1()
        self.nmos_3.add_edges(ckt, Ln, Wn, scale, cox_dash, n_params, int_2, gnd, out_b, gnd)

    def update_cap_values(self, cap_values, edge_dict, edge_voltages):
        cap_values = self.pmos_1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.pmos_2.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.pmos_3.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_2.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_3.update_cap_values(cap_values, edge_dict, edge_voltages)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        vccs_values = self.pmos_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_3.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_3.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values

###################
# NOR
###################
class NorDigital():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, Ln, Wn, Lp, Wp, scale, cox_dash, pwr, gnd, in_a, in_b, out_b):

        int_1 = ckt.get_internal_node()

        self.pmos_1 = PmosDigital()
        self.pmos_1.add_edges(ckt, Lp, Wp, scale, cox_dash, in_a, pwr, int_1)

        self.pmos_2 = PmosDigital()
        self.pmos_2.add_edges(ckt, Lp, Wp, scale, cox_dash, in_b, int_1, out_b)

        self.nmos_1 = NmosDigital()
        self.nmos_1.add_edges(ckt, Ln, Wn, scale, cox_dash, in_b, gnd, out_b)

        self.nmos_2 = NmosDigital()
        self.nmos_2.add_edges(ckt, Ln, Wn, scale, cox_dash, in_a, gnd, out_b)

    def update_cap_values(self, cap_values, edge_dict):
        cap_values = self.pmos_1.update_cap_values(cap_values, edge_dict)
        cap_values = self.pmos_2.update_cap_values(cap_values, edge_dict)
        cap_values = self.nmos_1.update_cap_values(cap_values, edge_dict)
        cap_values = self.nmos_2.update_cap_values(cap_values, edge_dict)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        vccs_values = self.pmos_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values

    def get_vccs_values_ref(self):

        vccs_values_ref = []

        vccs_values_ref.append(self.pmos_1.update_vccs_values)
        vccs_values_ref.append(self.pmos_2.update_vccs_values)
        vccs_values_ref.append(self.nmos_1.update_vccs_values)
        vccs_values_ref.append(self.nmos_2.update_vccs_values)

        return vccs_values_ref

###################
# NOR L1
###################
class NorDigitalL1():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, in_a, in_b, out_b):

        int_1 = ckt.get_internal_node()

        self.pmos_1 = PmosDigitalL1()
        self.pmos_1.add_edges(ckt, Lp, Wp, scale, cox_dash, p_params, in_a, pwr, int_1, pwr)

        self.pmos_2 = PmosDigitalL1()
        self.pmos_2.add_edges(ckt, Lp, Wp, scale, cox_dash, p_params, in_b, int_1, out_b, pwr)

        self.nmos_1 = NmosDigitalL1()
        self.nmos_1.add_edges(ckt, Ln, Wn, scale, cox_dash, n_params, in_b, gnd, out_b, gnd)

        self.nmos_2 = NmosDigitalL1()
        self.nmos_2.add_edges(ckt, Ln, Wn, scale, cox_dash, n_params, in_a, gnd, out_b, gnd)

    def update_cap_values(self, cap_values, edge_dict, edge_voltages):
        cap_values = self.pmos_1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.pmos_2.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_2.update_cap_values(cap_values, edge_dict, edge_voltages)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        vccs_values = self.pmos_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values

###################
# OR
###################
class OrDigital():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, Ln, Wn, Lp, Wp, scale, cox_dash, pwr, gnd, in_a, in_b, out_b):

        int_1 = ckt.get_internal_node()
        int_2 = ckt.get_internal_node()

        # NOR
        self.pmos_1 = PmosDigital()
        self.pmos_1.add_edges(ckt, Lp, Wp, scale, cox_dash, in_a, pwr, int_1)

        self.pmos_2 = PmosDigital()
        self.pmos_2.add_edges(ckt, Lp, Wp, scale, cox_dash, in_b, int_1, int_2)

        self.nmos_1 = NmosDigital()
        self.nmos_1.add_edges(ckt, Ln, Wn, scale, cox_dash, in_b, gnd, int_2)

        self.nmos_2 = NmosDigital()
        self.nmos_2.add_edges(ckt, Ln, Wn, scale, cox_dash, in_a, gnd, int_2)

        # Inverter
        self.pmos_3 = PmosDigital()
        self.pmos_3.add_edges(ckt, Lp, Wp, scale, cox_dash, int_2, pwr, out_b)

        self.nmos_3 = NmosDigital()
        self.nmos_3.add_edges(ckt, Ln, Wn, scale, cox_dash, int_2, gnd, out_b)

    def update_cap_values(self, cap_values, edge_dict):
        cap_values = self.pmos_1.update_cap_values(cap_values, edge_dict)
        cap_values = self.pmos_2.update_cap_values(cap_values, edge_dict)
        cap_values = self.pmos_3.update_cap_values(cap_values, edge_dict)
        cap_values = self.nmos_1.update_cap_values(cap_values, edge_dict)
        cap_values = self.nmos_2.update_cap_values(cap_values, edge_dict)
        cap_values = self.nmos_3.update_cap_values(cap_values, edge_dict)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        vccs_values = self.pmos_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_3.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_3.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values

    def get_vccs_values_ref(self):

        vccs_values_ref = []

        vccs_values_ref.append(self.pmos_1.update_vccs_values)
        vccs_values_ref.append(self.pmos_2.update_vccs_values)
        vccs_values_ref.append(self.pmos_3.update_vccs_values)
        vccs_values_ref.append(self.nmos_1.update_vccs_values)
        vccs_values_ref.append(self.nmos_2.update_vccs_values)
        vccs_values_ref.append(self.nmos_3.update_vccs_values)

        return vccs_values_ref

###################
# OR L1
###################
class OrDigitalL1():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, in_a, in_b, out_b):

        int_1 = ckt.get_internal_node()
        int_2 = ckt.get_internal_node()

        # NOR
        self.pmos_1 = PmosDigitalL1()
        self.pmos_1.add_edges(ckt, Lp, Wp, scale, cox_dash, p_params, in_a, pwr, int_1, pwr)

        self.pmos_2 = PmosDigitalL1()
        self.pmos_2.add_edges(ckt, Lp, Wp, scale, cox_dash, p_params, in_b, int_1, int_2, pwr)

        self.nmos_1 = NmosDigitalL1()
        self.nmos_1.add_edges(ckt, Ln, Wn, scale, cox_dash, n_params, in_b, gnd, int_2, gnd)

        self.nmos_2 = NmosDigitalL1()
        self.nmos_2.add_edges(ckt, Ln, Wn, scale, cox_dash, n_params, in_a, gnd, int_2, gnd)

        # Inverter
        self.pmos_3 = PmosDigitalL1()
        self.pmos_3.add_edges(ckt, Lp, Wp, scale, cox_dash, p_params, int_2, pwr, out_b, pwr)

        self.nmos_3 = NmosDigitalL1()
        self.nmos_3.add_edges(ckt, Ln, Wn, scale, cox_dash, n_params, int_2, gnd, out_b, gnd)

    def update_cap_values(self, cap_values, edge_dict, edge_voltages):
        cap_values = self.pmos_1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.pmos_2.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.pmos_3.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_2.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_3.update_cap_values(cap_values, edge_dict, edge_voltages)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        vccs_values = self.pmos_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_3.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_3.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values

###################
# XOR
###################
class XorDigital():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, Ln, Wn, Lp, Wp, scale, cox_dash, pwr, gnd, in_a, in_b, out_b):

        int_1 = ckt.get_internal_node()
        int_2 = ckt.get_internal_node()

        self.nand_1 = NandDigital()
        self.nand_1.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, pwr, gnd, in_a, in_b, int_1)

        self.or_1 = OrDigital()
        self.or_1.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, pwr, gnd, in_a, in_b, int_2)

        self.and_1 = AndDigital()
        self.and_1.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, pwr, gnd, int_1, int_2, out_b)

    def update_cap_values(self, cap_values, edge_dict):
        cap_values = self.nand_1.update_cap_values(cap_values, edge_dict)
        cap_values = self.or_1.update_cap_values(cap_values, edge_dict)
        cap_values = self.and_1.update_cap_values(cap_values, edge_dict)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        vccs_values = self.nand_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.or_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.and_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values

    def get_vccs_values_ref(self):

        vccs_values_ref = []

        refs_nand_1 = self.nand_1.get_vccs_values_ref()
        for i in range(len(refs_nand_1)):
            vccs_values_ref.append(refs_nand_1[i])

        refs_or_1 = self.or_1.get_vccs_values_ref()
        for i in range(len(refs_or_1)):
            vccs_values_ref.append(refs_or_1[i])

        refs_and_1 = self.and_1.get_vccs_values_ref()
        for i in range(len(refs_and_1)):
            vccs_values_ref.append(refs_and_1[i])

        return vccs_values_ref

###################
# XOR L1
###################
class XorDigitalL1():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, in_a, in_b, out_b):

        int_1 = ckt.get_internal_node()
        int_2 = ckt.get_internal_node()

        self.nand_1 = NandDigitalL1()
        self.nand_1.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, in_a, in_b, int_1)

        self.or_1 = OrDigitalL1()
        self.or_1.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, in_a, in_b, int_2)

        self.and_1 = AndDigitalL1()
        self.and_1.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, int_1, int_2, out_b)

    def update_cap_values(self, cap_values, edge_dict, edge_voltages):
        cap_values = self.nand_1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.or_1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.and_1.update_cap_values(cap_values, edge_dict, edge_voltages)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        vccs_values = self.nand_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.or_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.and_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values

###################
# FULL ADDER
###################
class FullAdderDigital():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, in_a, in_b, in_cin, out_sum, out_cout):

        int_1 = ckt.get_internal_node()
        int_2 = ckt.get_internal_node()
        int_3 = ckt.get_internal_node()

        self.xor_1 = XorDigital()
        self.xor_1.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, pwr, gnd, in_a, in_b, int_1)

        self.xor_2 = XorDigital()
        self.xor_2.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, pwr, gnd, int_1, in_cin, out_sum)

        self.and_1 = AndDigital()
        self.and_1.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, pwr, gnd, int_1, in_cin, int_2)

        self.and_2 = AndDigital()
        self.and_2.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, pwr, gnd, in_a, in_b, int_3)

        self.or_1 = OrDigital()
        self.or_1.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, pwr, gnd, int_2, int_3, out_cout)

    def update_cap_values(self, cap_values, edge_dict, edge_voltages):
        cap_values = self.xor_1.update_cap_values(cap_values, edge_dict)
        cap_values = self.xor_2.update_cap_values(cap_values, edge_dict)
        cap_values = self.and_1.update_cap_values(cap_values, edge_dict)
        cap_values = self.and_2.update_cap_values(cap_values, edge_dict)
        cap_values = self.or_1.update_cap_values(cap_values, edge_dict)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        vccs_values = self.xor_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.xor_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.and_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.and_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.or_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values

    def get_vccs_values_ref(self):

        vccs_values_ref = []

        refs_xor_1 = self.xor_1.get_vccs_values_ref()
        for i in range(len(refs_xor_1)):
            vccs_values_ref.append(refs_xor_1[i])

        refs_xor_2 = self.xor_2.get_vccs_values_ref()
        for i in range(len(refs_xor_2)):
            vccs_values_ref.append(refs_xor_2[i])

        refs_and_1 = self.and_1.get_vccs_values_ref()
        for i in range(len(refs_and_1)):
            vccs_values_ref.append(refs_and_1[i])

        refs_and_2 = self.and_2.get_vccs_values_ref()
        for i in range(len(refs_and_2)):
            vccs_values_ref.append(refs_and_2[i])

        refs_or_1 = self.or_1.get_vccs_values_ref()
        for i in range(len(refs_or_1)):
            vccs_values_ref.append(refs_or_1[i])

        return vccs_values_ref

###################
# FULL ADDER L1
###################
class FullAdderDigitalL1():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, in_a, in_b, in_cin, out_sum, out_cout):

        int_1 = ckt.get_internal_node()
        int_2 = ckt.get_internal_node()
        int_3 = ckt.get_internal_node()

        self.xor_1 = XorDigitalL1()
        self.xor_1.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, in_a, in_b, int_1)

        self.xor_2 = XorDigitalL1()
        self.xor_2.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, int_1, in_cin, out_sum)

        self.and_1 = AndDigitalL1()
        self.and_1.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, int_1, in_cin, int_2)

        self.and_2 = AndDigitalL1()
        self.and_2.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, in_a, in_b, int_3)

        self.or_1 = OrDigitalL1()
        self.or_1.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, int_2, int_3, out_cout)

    def update_cap_values(self, cap_values, edge_dict, edge_voltages):
        cap_values = self.xor_1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.xor_2.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.and_1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.and_2.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.or_1.update_cap_values(cap_values, edge_dict, edge_voltages)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        vccs_values = self.xor_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.xor_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.and_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.and_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.or_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values

###################
# Transmission Gate
###################
class TGDigital():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, Ln, Wn, Lp, Wp, scale, cox_dash, in_a, out_a, clk, clk_n):

        int_1 = ckt.get_internal_node()

        self.pmos_1 = PmosDigital()
        self.pmos_1.add_edges(ckt, Lp, Wp, scale, cox_dash, clk_n, out_a, in_a)

        self.nmos_1 = NmosDigital()
        self.nmos_1.add_edges(ckt, Ln, Wn, scale, cox_dash, clk, out_a, in_a)

    def update_cap_values(self, cap_values, edge_dict, edge_voltages):
        cap_values = self.pmos_1.update_cap_values(cap_values, edge_dict)
        cap_values = self.nmos_1.update_cap_values(cap_values, edge_dict)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        vccs_values = self.pmos_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values

    def get_vccs_values_ref(self):

        vccs_values_ref = []

        vccs_values_ref.append(self.pmos_1.update_vccs_values)
        vccs_values_ref.append(self.nmos_1.update_vccs_values)

        return vccs_values_ref

###################
# Transmission Gate L1
###################
class TGDigitalL1():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, in_a, out_a, clk, clk_n):

        int_1 = ckt.get_internal_node()

        self.pmos_1 = PmosDigitalL1()
        self.pmos_1.add_edges(ckt, Lp, Wp, scale, cox_dash, p_params, clk_n, out_a, in_a, pwr)

        self.nmos_1 = NmosDigitalL1()
        self.nmos_1.add_edges(ckt, Ln, Wn, scale, cox_dash, n_params, clk, out_a, in_a, gnd)

    def update_cap_values(self, cap_values, edge_dict, edge_voltages):
        cap_values = self.pmos_1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_1.update_cap_values(cap_values, edge_dict, edge_voltages)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        vccs_values = self.pmos_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values

###################
# Inverter
###################
class InverterDigital():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, Ln, Wn, Lp, Wp, scale, cox_dash, pwr, gnd, in_a, out_a):

        self.pmos = PmosDigital()
        self.pmos.add_edges(ckt, Lp, Wp, scale, cox_dash, in_a, pwr, out_a)

        self.nmos = NmosDigital()
        self.nmos.add_edges(ckt, Ln, Wn, scale, cox_dash, in_a, gnd, out_a)

    def update_cap_values(self, cap_values, edge_dict, edge_voltages):
        cap_values = self.pmos.update_cap_values(cap_values, edge_dict)
        cap_values = self.nmos.update_cap_values(cap_values, edge_dict)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        vccs_values = self.pmos.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values

    def get_vccs_values_ref(self):

        vccs_values_ref = []

        vccs_values_ref.append(self.pmos.update_vccs_values)
        vccs_values_ref.append(self.nmos.update_vccs_values)

        return vccs_values_ref

###################
# Inverter L1
###################
class InverterDigitalL1():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, in_a, out_a):

        self.pmos = PmosDigitalL1()
        self.pmos.add_edges(ckt, Lp, Wp, scale, cox_dash, p_params, in_a, pwr, out_a, pwr)

        self.nmos = NmosDigitalL1()
        self.nmos.add_edges(ckt, Ln, Wn, scale, cox_dash, n_params, in_a, gnd, out_a, gnd)

    def update_cap_values(self, cap_values, edge_dict, edge_voltages):
        cap_values = self.pmos.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos.update_cap_values(cap_values, edge_dict, edge_voltages)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        vccs_values = self.pmos.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values

    def dy_update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages, dy_ref):
        if self.pmos.is_dy_dep(dy_ref) == True:
            vccs_values = self.pmos.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        if self.nmos.is_dy_dep(dy_ref) == True:
            vccs_values = self.nmos.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values

    def get_vccs_values(self, vccs_values):
        vccs_values = self.pmos.get_vccs_values(vccs_values)
        vccs_values = self.nmos.get_vccs_values(vccs_values)

        return vccs_values

    def update_vccs_mask(self, vccs_mask):
        vccs_mask = self.pmos.update_vccs_mask(vccs_mask)
        vccs_mask = self.nmos.update_vccs_mask(vccs_mask)

        return vccs_mask

###################
# Inverter BSIM
###################
class InverterBSIM():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, pwr, gnd, in_a, out_a, nmos_params, pmos_params, bdf_method):

        self.pmos = bsim_cmg(pmos_params)

        if pmos_params.get('RGATEMOD') != None:
            if pmos_params['RGATEMOD'] != 0:
               self.pmos.define_RGATEMOD()

        self.pmos.add_edges(ckt, in_a, pwr, out_a, pwr)

        self.nmos = bsim_cmg(nmos_params)

        if nmos_params.get('RGATEMOD') != None:
            if nmos_params['RGATEMOD'] != 0:
               self.nmos.define_RGATEMOD()

        self.nmos.add_edges(ckt, in_a, gnd, out_a, gnd)

        self.pmos.CMGBiasIndepCalc()
        self.nmos.CMGBiasIndepCalc()

        self.pmos.set_ddt(bdf_method)
        self.nmos.set_ddt(bdf_method)

    def update_cap_values(self, cap_values, edge_dict, edge_voltages):
        cap_values = self.pmos.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos.update_cap_values(cap_values, edge_dict, edge_voltages)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        try:
            vccs_values = self.pmos.CMGBiasDepCalc(vccs_values, edge_voltages, t)
            vccs_values = self.nmos.CMGBiasDepCalc(vccs_values, edge_voltages, t)
        except:
            #print (edge_voltages)
            #sys.exit(0)
            #vccs_values = vccs_values
            vccs_values.fill(-1)

        return vccs_values

    def get_vccs_values(self, vccs_values):
        vccs_values = self.pmos.get_vccs_values(vccs_values)
        vccs_values = self.nmos.get_vccs_values(vccs_values)

        return vccs_values

    def update_vccs_mask(self, vccs_mask):
        vccs_mask = self.pmos.update_vccs_mask(vccs_mask)
        vccs_mask = self.nmos.update_vccs_mask(vccs_mask)

        return vccs_mask

    def update_charge(self):
        self.pmos.update_charge()
        self.nmos.update_charge()

###################
# AND BSIM
###################
class ANDBSIM():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, pwr, gnd, in_a, in_b, out_b, nmos_params, pmos_params, bdf_method):

        pmos_params['L'] = 2 * pmos_params['L']

        int_1 = ckt.get_internal_node()
        int_2 = ckt.get_internal_node()

        # NAND
        self.pmos_1 = bsim_cmg(pmos_params)
        self.pmos_2 = bsim_cmg(pmos_params)
        self.nmos_1 = bsim_cmg(nmos_params)
        self.nmos_2 = bsim_cmg(nmos_params)

        # Inverter
        self.pmos_3 = bsim_cmg(pmos_params)
        self.nmos_3 = bsim_cmg(nmos_params)

        if pmos_params.get('RGATEMOD') != None:
            if pmos_params['RGATEMOD'] != 0:
               self.pmos_1.define_RGATEMOD()
               self.pmos_2.define_RGATEMOD()
               self.pmos_3.define_RGATEMOD()

        if nmos_params.get('RGATEMOD') != None:
            if nmos_params['RGATEMOD'] != 0:
               self.nmos_1.define_RGATEMOD()
               self.nmos_2.define_RGATEMOD()
               self.nmos_3.define_RGATEMOD()

        # NAND
        self.pmos_1.add_edges(ckt, in_a, pwr, int_2, pwr)
        self.pmos_2.add_edges(ckt, in_b, pwr, int_2, pwr)
        self.nmos_1.add_edges(ckt, in_a, int_1, int_2, gnd)
        self.nmos_2.add_edges(ckt, in_b, gnd, int_1, gnd)

        # Inverter
        self.pmos_3.add_edges(ckt, int_2, pwr, out_b, pwr)
        self.nmos_3.add_edges(ckt, int_2, gnd, out_b, gnd)

        self.pmos_1.CMGBiasIndepCalc()
        self.pmos_2.CMGBiasIndepCalc()
        self.pmos_3.CMGBiasIndepCalc()
        self.nmos_1.CMGBiasIndepCalc()
        self.nmos_2.CMGBiasIndepCalc()
        self.nmos_3.CMGBiasIndepCalc()

        self.pmos_1.set_ddt(bdf_method)
        self.pmos_2.set_ddt(bdf_method)
        self.pmos_3.set_ddt(bdf_method)
        self.nmos_1.set_ddt(bdf_method)
        self.nmos_2.set_ddt(bdf_method)
        self.nmos_3.set_ddt(bdf_method)

    def update_cap_values(self, cap_values, edge_dict, edge_voltages):
        cap_values = self.pmos.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos.update_cap_values(cap_values, edge_dict, edge_voltages)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        try:
            vccs_values = self.pmos_1.CMGBiasDepCalc(vccs_values, edge_voltages, t)
            vccs_values = self.pmos_2.CMGBiasDepCalc(vccs_values, edge_voltages, t)
            vccs_values = self.pmos_3.CMGBiasDepCalc(vccs_values, edge_voltages, t)
            vccs_values = self.nmos_1.CMGBiasDepCalc(vccs_values, edge_voltages, t)
            vccs_values = self.nmos_2.CMGBiasDepCalc(vccs_values, edge_voltages, t)
            vccs_values = self.nmos_3.CMGBiasDepCalc(vccs_values, edge_voltages, t)
        except:
            #vccs_values = vccs_values
            vccs_values.fill(-1000)

        return vccs_values

    def get_vccs_values(self, vccs_values):
        vccs_values = self.pmos_1.get_vccs_values(vccs_values)
        vccs_values = self.pmos_2.get_vccs_values(vccs_values)
        vccs_values = self.pmos_3.get_vccs_values(vccs_values)
        vccs_values = self.nmos_1.get_vccs_values(vccs_values)
        vccs_values = self.nmos_2.get_vccs_values(vccs_values)
        vccs_values = self.nmos_3.get_vccs_values(vccs_values)

        return vccs_values

    def update_vccs_mask(self, vccs_mask):
        vccs_mask = self.pmos_1.update_vccs_mask(vccs_mask)
        vccs_mask = self.pmos_2.update_vccs_mask(vccs_mask)
        vccs_mask = self.pmos_3.update_vccs_mask(vccs_mask)
        vccs_mask = self.nmos_1.update_vccs_mask(vccs_mask)
        vccs_mask = self.nmos_2.update_vccs_mask(vccs_mask)
        vccs_mask = self.nmos_3.update_vccs_mask(vccs_mask)

        return vccs_mask

    def update_charge(self):
        self.pmos_1.update_charge()
        self.pmos_2.update_charge()
        self.pmos_3.update_charge()
        self.nmos_1.update_charge()
        self.nmos_2.update_charge()
        self.nmos_3.update_charge()

###################
# Flip Flop
###################
class FlipFlopDigital():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, Ln, Wn, Lp, Wp, scale, cox_dash, pwr, gnd, d, q, clk, clk_n):

        int_1 = ckt.get_internal_node()
        int_2 = ckt.get_internal_node()
        int_3 = ckt.get_internal_node()
        int_4 = ckt.get_internal_node()
        int_5 = ckt.get_internal_node()

        Wn_tg = Wn
        Wp_tg = Wp

        # Master
        self.tg_1 = TGDigital()
        self.tg_1.add_edges(ckt, Ln, Wn_tg, Lp, Wp_tg, scale, cox_dash, d, int_1, clk_n, clk)

        self.inv_1 = InverterDigital()
        self.inv_1.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, pwr, gnd, int_1, int_2)

        self.inv_2 = InverterDigital()
        self.inv_2.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, pwr, gnd, int_2, int_3)

        self.tg_2 = TGDigital()
        self.tg_2.add_edges(ckt, Ln, Wn_tg, Lp, Wp_tg, scale, cox_dash, int_3, int_1, clk, clk_n)

        # Slave
        self.tg_3 = TGDigital()
        self.tg_3.add_edges(ckt, Ln, Wn_tg, Lp, Wp_tg, scale, cox_dash, int_2, int_4, clk, clk_n)

        self.inv_3 = InverterDigital()
        self.inv_3.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, pwr, gnd, int_4, q)

        self.inv_4 = InverterDigital()
        self.inv_4.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, pwr, gnd, q, int_5)

        self.tg_4 = TGDigital()
        self.tg_4.add_edges(ckt, Ln, Wn_tg, Lp, Wp_tg, scale, cox_dash, int_5, int_4, clk_n, clk)

    def update_cap_values(self, cap_values, edge_dict, edge_voltages):
        cap_values = self.tg_1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.tg_2.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.tg_3.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.tg_4.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.inv_1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.inv_2.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.inv_3.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.inv_4.update_cap_values(cap_values, edge_dict, edge_voltages)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        vccs_values = self.tg_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.tg_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.tg_3.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.tg_4.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.inv_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.inv_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.inv_3.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.inv_4.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values

###################
# Flip Flop L1
###################
class FlipFlopDigitalL1():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, d, q, clk, clk_n):

        int_1 = ckt.get_internal_node()
        int_2 = ckt.get_internal_node()
        int_3 = ckt.get_internal_node()
        int_4 = ckt.get_internal_node()
        int_5 = ckt.get_internal_node()

        Wn_tg = Wn
        Wp_tg = Wp

        # Master
        self.tg_1 = TGDigitalL1()
        self.tg_1.add_edges(ckt, Ln, Wn_tg, Lp, Wp_tg, scale, cox_dash, n_params, p_params, pwr, gnd, d, int_1, clk_n, clk)

        self.inv_1 = InverterDigitalL1()
        self.inv_1.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, int_1, int_2)

        self.inv_2 = InverterDigitalL1()
        self.inv_2.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, int_2, int_3)

        self.tg_2 = TGDigitalL1()
        self.tg_2.add_edges(ckt, Ln, Wn_tg, Lp, Wp_tg, scale, cox_dash, n_params, p_params, pwr, gnd, int_3, int_1, clk, clk_n)

        # Slave
        self.tg_3 = TGDigitalL1()
        self.tg_3.add_edges(ckt, Ln, Wn_tg, Lp, Wp_tg, scale, cox_dash, n_params, p_params, pwr, gnd, int_2, int_4, clk, clk_n)

        self.inv_3 = InverterDigitalL1()
        self.inv_3.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, int_4, q)

        self.inv_4 = InverterDigitalL1()
        self.inv_4.add_edges(ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, q, int_5)

        self.tg_4 = TGDigitalL1()
        self.tg_4.add_edges(ckt, Ln, Wn_tg, Lp, Wp_tg, scale, cox_dash, n_params, p_params, pwr, gnd, int_5, int_4, clk_n, clk)

    def update_cap_values(self, cap_values, edge_dict, edge_voltages):
        cap_values = self.tg_1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.tg_2.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.tg_3.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.tg_4.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.inv_1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.inv_2.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.inv_3.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.inv_4.update_cap_values(cap_values, edge_dict, edge_voltages)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        vccs_values = self.tg_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.tg_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.tg_3.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.tg_4.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.inv_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.inv_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.inv_3.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.inv_4.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values


###################
# Sense Amplifier L0
###################
class SenseAmp():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, in_p, in_n, out_p, out_n, clk):

        int_3 = ckt.get_internal_node()

        self.pmos_m3 = PmosDigital() # g, s, d
        self.pmos_m3.add_edges(ckt, 1, 20, scale, cox_dash,  out_n, pwr, out_p)

        self.nmos_m1 = NmosDigital() # g, s, d
        self.nmos_m1.add_edges(ckt, 1, 10, scale, cox_dash,  out_n, int_3, out_p)

        self.pmos_m4 = PmosDigital() # g, s, d
        self.pmos_m4.add_edges(ckt, 1, 20, scale, cox_dash,  out_p, pwr, out_n)

        self.nmos_m2 = NmosDigital() # g, s, d
        self.nmos_m2.add_edges(ckt, 1, 10, scale, cox_dash,  out_p, int_3, out_n)

        self.pmos_ms1 = PmosDigital() # g, s, d
        self.pmos_ms1.add_edges(ckt, 1, 20, scale, cox_dash, clk, in_p, out_p)

        self.pmos_ms2 = PmosDigital() # g, s, d
        self.pmos_ms2.add_edges(ckt, 1, 20, scale, cox_dash, clk, in_n, out_n)

        self.nmos_ms3 = NmosDigital() # g, s, d
        self.nmos_ms3.add_edges(ckt, 1, 10, scale, cox_dash, clk, gnd, int_3)

    def update_cap_values(self, cap_values, edge_dict, edge_voltages):
        cap_values = self.pmos_m3.update_cap_values(cap_values, edge_dict)
        cap_values = self.pmos_m4.update_cap_values(cap_values, edge_dict)
        cap_values = self.pmos_ms1.update_cap_values(cap_values, edge_dict)
        cap_values = self.pmos_ms2.update_cap_values(cap_values, edge_dict)

        cap_values = self.nmos_m1.update_cap_values(cap_values, edge_dict)
        cap_values = self.nmos_m2.update_cap_values(cap_values, edge_dict)
        cap_values = self.nmos_ms3.update_cap_values(cap_values, edge_dict)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        vccs_values = self.pmos_m3.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_m4.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_ms1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_ms2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        vccs_values = self.nmos_m1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_m2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_ms3.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values

###################
# Sense Amplifier L1
###################

# Note : This circuit does not function if Vthp exceeds input voltages (noted in CMOS text).
# To prevent such problems, the GAMMA parameter is kept small
class SenseAmpL1():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, in_p, in_n, out_p, out_n, clk):

        int_3 = ckt.get_internal_node()

        self.pmos_m3 = PmosDigitalL1() # g, s, d, b
        self.pmos_m3.add_edges(ckt, 1, 20, scale, cox_dash, p_params, out_n, pwr, out_p, pwr)

        self.nmos_m1 = NmosDigitalL1() # g, s, d, b
        self.nmos_m1.add_edges(ckt, 1, 10, scale, cox_dash, n_params, out_n, int_3, out_p, gnd)

        self.pmos_m4 = PmosDigitalL1() # g, s, d, b
        self.pmos_m4.add_edges(ckt, 1, 20, scale, cox_dash, p_params, out_p, pwr, out_n, pwr)

        self.nmos_m2 = NmosDigitalL1() # g, s, d, b
        self.nmos_m2.add_edges(ckt, 1, 10, scale, cox_dash, n_params, out_p, int_3, out_n, gnd)

        # The source and drain of ms1 are swapped here w.r.t CMOS textbook. It appears that the PMOS
        # source and drain terminals are not interchangable in the current model.
        self.pmos_ms1 = PmosDigitalL1() # g, s, d, b
        self.pmos_ms1.add_edges(ckt, 1, 20, scale, cox_dash, p_params, clk, in_p, out_p, pwr)

        self.pmos_ms2 = PmosDigitalL1() # g, s, d, b
        self.pmos_ms2.add_edges(ckt, 1, 20, scale, cox_dash, p_params, clk, in_n, out_n, pwr)

        self.nmos_ms3 = NmosDigitalL1() # g, s, d, b
        self.nmos_ms3.add_edges(ckt, 1, 10, scale, cox_dash, n_params, clk, gnd, int_3, gnd)

    def update_cap_values(self, cap_values, edge_dict, edge_voltages):
        cap_values = self.pmos_m3.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.pmos_m4.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.pmos_ms1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.pmos_ms2.update_cap_values(cap_values, edge_dict, edge_voltages)

        cap_values = self.nmos_m1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_m2.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_ms3.update_cap_values(cap_values, edge_dict, edge_voltages)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        vccs_values = self.pmos_m3.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_m4.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_ms1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_ms2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        vccs_values = self.nmos_m1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_m2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_ms3.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values

###################
# Sense Amplifier L1 ALT
###################

class SenseAmpL1ALT():

    def __init__(self):
        self.ref = 0

    def add_edges(self, ckt, Ln, Wn, Lp, Wp, scale, cox_dash, n_params, p_params, pwr, gnd, in_p, in_n, q, q_n, clk):

        int_1 = ckt.get_internal_node()
        int_2 = ckt.get_internal_node()
        int_3 = ckt.get_internal_node()
        int_4 = ckt.get_internal_node()
        int_5 = ckt.get_internal_node()
        int_6 = ckt.get_internal_node()

        int_op = ckt.get_internal_node()
        int_on = ckt.get_internal_node()

        self.pmos_ms3 = PmosDigitalL1() # g, s, d, b
        self.pmos_ms3.add_edges(ckt, 1, 20, scale, cox_dash, p_params, clk, pwr, int_on, pwr)

        self.pmos_m3 = PmosDigitalL1() # g, s, d, b
        self.pmos_m3.add_edges(ckt, 1, 20, scale, cox_dash, p_params, int_op, pwr, int_on, pwr)

        self.pmos_ms4 = PmosDigitalL1() # g, s, d, b
        self.pmos_ms4.add_edges(ckt, 1, 20, scale, cox_dash, p_params, clk, pwr, int_op, pwr)

        self.pmos_m4 = PmosDigitalL1() # g, s, d, b
        self.pmos_m4.add_edges(ckt, 1, 20, scale, cox_dash, p_params, int_on, pwr, int_op, pwr)

        self.nmos_ms1 = NmosDigitalL1() # g, s, d, b
        self.nmos_ms1.add_edges(ckt, 1, 10, scale, cox_dash, n_params, clk, int_1, int_on, gnd)

        self.nmos_ms2 = NmosDigitalL1() # g, s, d, b
        self.nmos_ms2.add_edges(ckt, 1, 10, scale, cox_dash, n_params, clk, int_2, int_op, gnd)

        self.pmos_mb7 = PmosDigitalL1() # g, s, d, b
        self.pmos_mb7.add_edges(ckt, 1, 20, scale, cox_dash, p_params, in_n, pwr, int_3, pwr)

        self.nmos_mb5 = NmosDigitalL1() # g, s, d, b
        self.nmos_mb5.add_edges(ckt, 1, 10, scale, cox_dash, n_params, int_3, gnd, int_3, gnd)

        self.nmos_mb3 = NmosDigitalL1() # g, s, d, b
        self.nmos_mb3.add_edges(ckt, 1, 10, scale, cox_dash, n_params, int_3, gnd, int_4, gnd)

        self.nmos_mb1 = NmosDigitalL1() # g, s, d, b
        self.nmos_mb1.add_edges(ckt, 1, 10, scale, cox_dash, n_params, in_p, gnd, int_4, gnd)

        self.nmos_m1 = NmosDigitalL1() # g, s, d, b
        self.nmos_m1.add_edges(ckt, 1, 10, scale, cox_dash, n_params, int_op, int_4, int_1, gnd)

        self.pmos_mb8 = PmosDigitalL1() # g, s, d, b
        self.pmos_mb8.add_edges(ckt, 1, 20, scale, cox_dash, p_params, in_p, pwr, int_5, pwr)

        self.nmos_mb6 = NmosDigitalL1() # g, s, d, b
        self.nmos_mb6.add_edges(ckt, 1, 10, scale, cox_dash, n_params, int_5, gnd, int_5, gnd)

        self.nmos_mb4 = NmosDigitalL1() # g, s, d, b
        self.nmos_mb4.add_edges(ckt, 1, 10, scale, cox_dash, n_params, int_5, gnd, int_6, gnd)

        self.nmos_mb2 = NmosDigitalL1() # g, s, d, b
        self.nmos_mb2.add_edges(ckt, 1, 10, scale, cox_dash, n_params, in_n, gnd, int_6, gnd)

        self.nmos_m2 = NmosDigitalL1() # g, s, d, b
        self.nmos_m2.add_edges(ckt, 1, 10, scale, cox_dash, n_params, int_on, int_6, int_2, gnd)

        self.nand_1 = NandDigitalL1() # pwr, gnd, in_a, in_b, out_b
        self.nand_1.add_edges(ckt, 1, 10, 1, 20, scale, cox_dash, n_params, p_params, pwr, gnd, int_op, q, q_n)

        self.nand_2 = NandDigitalL1() # pwr, gnd, in_a, in_b, out_b
        self.nand_2.add_edges(ckt, 1, 10, 1, 20, scale, cox_dash, n_params, p_params, pwr, gnd, int_on, q_n, q)

    def update_cap_values(self, cap_values, edge_dict, edge_voltages):
        cap_values = self.pmos_m3.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.pmos_m4.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.pmos_ms3.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.pmos_ms4.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.pmos_mb7.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.pmos_mb8.update_cap_values(cap_values, edge_dict, edge_voltages)

        cap_values = self.nmos_ms1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_ms2.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_mb5.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_mb3.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_mb1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_m1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_mb2.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_mb4.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_mb6.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nmos_m2.update_cap_values(cap_values, edge_dict, edge_voltages)

        cap_values = self.nand_1.update_cap_values(cap_values, edge_dict, edge_voltages)
        cap_values = self.nand_2.update_cap_values(cap_values, edge_dict, edge_voltages)

        return cap_values

    def update_vccs_values(self, vccs_values, edge_dict, t, edge_voltages):
        vccs_values = self.pmos_m3.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_m4.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_ms3.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_ms4.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_mb7.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.pmos_mb8.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        vccs_values = self.nmos_ms1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_ms2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_mb5.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_mb3.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_mb1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_m1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_mb2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_mb4.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_mb6.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nmos_m2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        vccs_values = self.nand_1.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)
        vccs_values = self.nand_2.update_vccs_values(vccs_values, edge_dict, t, edge_voltages)

        return vccs_values
