from enum import Enum
import sys, getopt, re, shutil, os, pprint, networkx as nx
import numpy as np

class ElementType(Enum):
    undefined   = 0
    resistor    = 1
    capacitor   = 2
    inductor    = 3
    voltage_src = 4
    current_src = 5

class TwoPortElement:

    def __init__(self):
        self.value     = 0.0
        self.type      = ElementType.undefined
        self.dynamic   = False
        self.dependant = False

    def set_value(self, value):
        self.value = value

    def set_type(self, type):
        self.type = type

    def set_dynamic(self, dynamic):
        self.dynamic = dynamic

    def set_dependant(self, dependant):
        self.dependant = dependant

    def get_type(self):
        return self.type

    def get_value(self, host_circuit, t, voltages):
        return self.value

    def get_dy_value(self, host_circuit, t, voltages):
        return 0.0

    def set_instance(self, instance):
        self.instance = instance

    def get_instance(self):
        return self.instance

    def set_ref(self, ref):
        self.ref = ref

    def get_ref(self):
        return self.ref

    def get_dynamic(self):
        return self.dynamic

    def get_dependant(self):
        return self.dependant

class Circuit:

    def __init__(self):
        self.num_edges = 0
        self.G = nx.MultiGraph()
        self.internal_node_count = -1

    # Get the current num_edges
    def get_num_edges(self):
        return self.num_edges

    def set_sys_vector(self, sys_var):
        self.sys_var = sys_var

    def set_time(self, t):
        self.t = t

    def get_internal_node(self):
        self.internal_node_count -= 1

        return self.internal_node_count + 1

    def get_num_sys_vars(self):
        return self.num_sys_var_cap + self.num_sys_var_ind

    def get_num_sys_var_cap(self):
        return self.num_sys_var_cap

    def get_num_non_sys_var_cap(self):
        return self.num_non_sys_var_cap

    def get_num_non_sys_var_ind(self):
        return self.num_non_sys_var_ind

    def get_num_sys_var_ind(self):
        return self.num_sys_var_ind

    def get_num_caps(self):
        return self.num_caps

    def get_num_inductors(self):
        return self.num_inductors

    def get_num_dep_src(self):
        return self.num_dep_srcs

    def get_edge_voltages(self):
        return self.edge_voltages

    # Add an edge to the Graph.
    # Note that explicit pos_node and neg_node attributes are stored as part of the edge dict
    # as edge iterators return a (u,v) node pair which are not always in the order of (pos,neg).
    # Note that each edge has a unique key. For some types of graph, an edge iterator does not return
    # the edge key. Therefore, store a copy in the dict (ref).
    def add_edge(self, pos_node, neg_node, edge_info):

        edge_type = edge_info.get_type()

        # Edge weights must be stored as part of the dict as this is used by the minimum_spanning_tree
        # method used to calculate the minimum weight spanning-tree.
        if edge_type == ElementType.resistor:
            edge_weight = 2
        elif edge_type == ElementType.capacitor:
            edge_weight = 0
        elif edge_type == ElementType.inductor:
            edge_weight = 3
        elif edge_type == ElementType.voltage_src:
            edge_weight = 0
        elif edge_type == ElementType.current_src:
            edge_weight = 4
        else:
            print ("No type specified") # REVISIT and throw exception

        edge_info.set_ref(self.num_edges)

        self.G.add_edge(pos_node,
                        neg_node,
                        key=self.num_edges,
                        weight=edge_weight,
                        pos_node=pos_node,
                        neg_node=neg_node,
                        ref=self.num_edges,
                        sys_var_ref=-1,
                        info=edge_info)
        self.num_edges += 1

        return self.num_edges - 1

    def consistency_check(self):

        consistency_failed = False

        if not nx.is_connected(self.G):
            print ("Warning: The graph is not connected")
            consistency_failed = True

        for node in self.G.nodes_iter():
            connected_nodes = 0
            for c_node in nx.all_neighbors(self.G,node):
                connected_nodes = connected_nodes + 1

            if connected_nodes < 2:
                print ("Warning: Node " + str(node) + " is only connected to node " + str(c_node))
                consistency_failed = True

        return consistency_failed

    def get_spanning_tree(self):
        self.spanning_tree = nx.minimum_spanning_tree(self.G)

    def get_co_tree(self):

        self.co_tree = self.G.copy()

        # Remove all edges that are in the spanning-tree from the co-tree
        for (u,v,d) in self.spanning_tree.edges(data=True):
            key = d['ref']
            self.co_tree.remove_edge(u,v,key)

    # ------------------------------------------
    # These methods determine edge attributes.
    # ------------------------------------------

    # Determine if a particular edge ref is a system variable capacitor.
    # That is, the capacitor is a member of the spanning-tree.
    def is_sys_var_cap(self, ref):

        for (u,v,d) in self.spanning_tree.edges(data=True):
            edge     = d['info']
            edge_ref = d['ref']

            if edge.get_type() == ElementType.capacitor and ref == edge_ref:
                return True

        return False

    # Determine if a particular edge ref is a system variable inductor.
    # That is, the inductor is a member of the co-tree.
    def is_sys_var_ind(self, ref):

        for (u,v,d) in self.co_tree.edges(data=True):
            edge     = d['info']
            edge_ref = d['ref']

            if edge.get_type() == ElementType.inductor and ref == edge_ref:
                return True

        return False

    # Determine if a particular edge ref is a voltage source.
    # Voltage sources are always present in the spanning-tree. If voltage
    # sources are connected in parallel, an exception is generated and therefore
    # it is assumed impossible to occur.
    def is_vsrc(self, ref):

        for (u,v,d) in self.spanning_tree.edges(data=True):
            edge     = d['info']
            edge_ref = d['ref']

            if edge.get_type() == ElementType.voltage_src and ref == edge_ref:
                return True

        return False

    # Determine if a particular edge ref is a current source.
    def is_isrc(self, ref):

        for (u,v,d) in self.co_tree.edges(data=True):
            edge     = d['info']
            edge_ref = d['ref']

            if edge.get_type() == ElementType.current_src and ref == edge_ref:
                return True

        for (u,v,d) in self.spanning_tree.edges(data=True):
            edge     = d['info']
            edge_ref = d['ref']

            if edge.get_type() == ElementType.current_src and ref == edge_ref:
                return True

        return False

    # Determine if a particular edge ref is a non-system variable capacitor.
    # That is, the capacitor is a member of the co-tree.
    def is_non_sys_var_cap(self, ref):

        for (u,v,d) in self.co_tree.edges(data=True):
            edge     = d['info']
            edge_ref = d['ref']

            if edge.get_type() == ElementType.capacitor and ref == edge_ref:
                return True

        return False

    # Determine if a particular edge ref is a non-system variable inductor.
    # That is, the inductor is a member of the spanning-tree.
    def is_non_sys_var_ind(self, ref):

        for (u,v,d) in self.spanning_tree.edges(data=True):
            edge     = d['info']
            edge_ref = d['ref']

            if edge.get_type() == ElementType.inductor and ref == edge_ref:
                return True

        return False

    # Determine if a particular edge ref is a resistor in the spanning-tree.
    def is_spanning_tree_res(self, ref):

        for (u,v,d) in self.spanning_tree.edges(data=True):
            edge     = d['info']
            edge_ref = d['ref']

            if edge.get_type() == ElementType.resistor and ref == edge_ref:
                return True

        return False

    # Determine if a particular edge ref is a resistor in the co-tree.
    def is_co_tree_res(self, ref):

        for (u,v,d) in self.co_tree.edges(data=True):
            edge     = d['info']
            edge_ref = d['ref']

            if edge.get_type() == ElementType.resistor and ref == edge_ref:
                return True

        return False

    def set_co_tree_res_list(self):

        self.edge_co_tree_res = []

        for (u,v,d) in self.co_tree.edges(data=True):
            edge     = d['info']
            edge_ref = d['ref']

            if edge.get_type() == ElementType.resistor:
                self.edge_co_tree_res.append(d)

    def get_co_tree_res_list(self):
        return self.edge_co_tree_res

    def set_spanning_tree_res_list(self):

        self.edge_spanning_tree_res = []

        for (u,v,d) in self.spanning_tree.edges(data=True):
            edge     = d['info']
            edge_ref = d['ref']

            if edge.get_type() == ElementType.resistor:
                self.edge_spanning_tree_res.append(d)

    def get_spanning_tree_res_list(self):
        return self.edge_spanning_tree_res

    def set_sys_var_cap_list(self):

        self.edge_sys_cap = []

        for (u,v,d) in self.spanning_tree.edges(data=True):

            edge = d['info']

            if edge.get_type()  == ElementType.capacitor:
                self.edge_sys_cap.append(d)

    def get_sys_var_cap_list(self):
        return self.edge_sys_cap

    def set_non_sys_var_cap_list(self):

        self.edge_non_sys_cap = []

        for (u,v,d) in self.co_tree.edges(data=True):

            edge = d['info']

            if edge.get_type()  == ElementType.capacitor:
                self.edge_non_sys_cap.append(d)

    def get_non_sys_var_cap_list(self):
        return self.edge_non_sys_cap

    def set_sys_var_ind_list(self):

        self.edge_sys_ind = []

        for (u,v,d) in self.co_tree.edges(data=True):

            edge = d['info']

            if edge.get_type()  == ElementType.inductor:
                self.edge_sys_ind.append(d)

    def set_non_sys_var_ind_list(self):

        self.edge_non_sys_ind = []

        for (u,v,d) in self.spanning_tree.edges(data=True):

            edge = d['info']

            if edge.get_type()  == ElementType.inductor:
                self.edge_non_sys_ind.append(d)

    def get_sys_var_ind_list(self):
        return self.edge_sys_ind

    def get_non_sys_var_ind_list(self):
        return self.edge_non_sys_ind

    def get_edge_info(self, ref):

        for (u,v,d) in self.spanning_tree.edges(data=True):
            edge     = d['info']
            edge_ref = d['ref']

            if ref == edge_ref:
                return edge

        for (u,v,d) in self.co_tree.edges(data=True):
            edge     = d['info']
            edge_ref = d['ref']

            if ref == edge_ref:
                return edge

        return -1 # REVIISIT is this OK?

    def get_edge_cap_ref(self, ref):

        for (u,v,d) in self.spanning_tree.edges(data=True):
            edge_ref = d['ref']

            if ref == edge_ref:
                cap_ref  = d['cap_ref']
                return cap_ref

        for (u,v,d) in self.co_tree.edges(data=True):
            edge_ref = d['ref']

            if ref == edge_ref:
                cap_ref  = d['cap_ref']
                return cap_ref

        return -1 # REVIISIT is this OK?

    def get_edge_ind_ref(self, ref):

        for (u,v,d) in self.spanning_tree.edges(data=True):
            edge_ref = d['ref']

            if ref == edge_ref:
                ind_ref  = d['ind_ref']
                return ind_ref

        for (u,v,d) in self.co_tree.edges(data=True):
            edge_ref = d['ref']

            if ref == edge_ref:
                ind_ref  = d['ind_ref']
                return ind_ref

        return -1 # REVIISIT is this OK?

    # Each edge has a unique ref. Capacitors and inductors that are members of the spanning-tree
    # and co-tree respectively are system variables. Assign these branches a unique sys_var_ref.
    # Since capacitors and inductors which are system variables are evaluated as part of distinct
    # matrices (Qf and Bf), the references both start at zero.
    def set_sys_var_ref(self):

        sys_var_cap_ref = 0
        sys_var_ind_ref = 0
        ind_ref         = 0

        for (u,v,d) in self.spanning_tree.edges(data=True):
            edge = d['info']

            if edge.get_type() == ElementType.capacitor:
                d['sys_var_ref'] = sys_var_cap_ref
                d['cap_ref']     = sys_var_cap_ref
                sys_var_cap_ref  += 1

            elif edge.get_type() == ElementType.inductor:
                d['ind_ref']     = ind_ref
                ind_ref          += 1

        cap_ref = sys_var_cap_ref

        for (u,v,d) in self.co_tree.edges(data=True):
            edge = d['info']

            if edge.get_type() == ElementType.inductor:
                d['sys_var_ref'] = sys_var_ind_ref
                d['ind_ref']     = ind_ref
                sys_var_ind_ref  += 1
                ind_ref          += 1

            elif edge.get_type() == ElementType.capacitor:
                d['cap_ref']     = cap_ref
                cap_ref          += 1

        self.num_sys_var_cap = sys_var_cap_ref
        self.num_sys_var_ind = sys_var_ind_ref
        self.num_caps        = cap_ref
        self.num_inductors   = ind_ref

        self.num_non_sys_var_cap = cap_ref - sys_var_cap_ref
        self.num_non_sys_var_ind = ind_ref - sys_var_ind_ref

    # Determine the sys_var_ref corresponding to a given ref. The dict entry for each edge defaults
    # to -1 and will remain as -1 unless the edge is a system variable.
    def get_sys_var_ref(self, ref):

        for (u,v,d) in self.spanning_tree.edges(data=True):
            sys_var_ref = d['sys_var_ref']
            edge_ref    = d['ref']

            if ref == edge_ref:
                return sys_var_ref

        for (u,v,d) in self.co_tree.edges(data=True):
            sys_var_ref = d['sys_var_ref']
            edge_ref    = d['ref']

            if ref == edge_ref:
                return sys_var_ref

        return -1

    def set_dep_var_ref(self):

        dep_ref = self.num_edges

        self.num_dep_srcs = 0
        self.edge_dep_src = []

        for (u,v,d) in self.spanning_tree.edges(data=True):
            edge = d['info']

            if edge.get_type() == ElementType.voltage_src and edge.get_dependant():
                d['dep_ref']       = dep_ref
                dep_ref           += 1
                self.num_dep_srcs += 1
                self.edge_dep_src.append(d)

            if edge.get_type() == ElementType.current_src and edge.get_dependant():
                d['dep_ref']       = dep_ref
                dep_ref           += 1
                self.num_dep_srcs += 1
                self.edge_dep_src.append(d)

        for (u,v,d) in self.co_tree.edges(data=True):
            edge = d['info']

            if edge.get_type() == ElementType.current_src and edge.get_dependant():
                d['dep_ref']       = dep_ref
                dep_ref           += 1
                self.num_dep_srcs += 1
                self.edge_dep_src.append(d)

    def get_dep_var_list(self):
        return self.edge_dep_src

    def set_src_edge_ref(self):

        self.src_edge = []

        for (u,v,d) in self.spanning_tree.edges(data=True):
            edge = d['info']

            if edge.get_type() == ElementType.voltage_src:
                self.src_edge.append(d)

            if edge.get_type() == ElementType.current_src:
                self.src_edge.append(d)

        for (u,v,d) in self.co_tree.edges(data=True):
            edge = d['info']

            if edge.get_type() == ElementType.current_src:
                self.src_edge.append(d)

    def get_src_edge_list(self):
        return self.src_edge

    # Calculates the Qf matrix relating to the network. The algorithm is as follows:
    # 1. For each edge of the spanning-tree, set Qf[ref][ref] = 1
    # 2. For the current spanning-tree edge, add a single co-tree edge.
    # 3. If the addition of the co-tree edge causes a loop to form, then the co-tree edge
    #    forms a basic cut with the current spanning-tree edge (see html description).
    # 4. Remove the added co-tree edge and current spanning-tree edge. The graph will now be
    #    disconnected and will be split into two sets of connected nodes.
    # 5. Determine which set of nodes contains the postive node of the spanning-tree edge.
    #    Positive current will be directed towards the postive node of the spanning-tree edge
    #    (this corresponds to the blue nodes in the html description). If the postive node of the
    #    co-tree edge is also contained in this set of nodes, the Qf entry is positive.
    def get_qf_matrix(self):

        self.qf = np.zeros([self.num_edges, self.num_edges])

        for (u,v,d) in self.spanning_tree.edges(data=True):

            spanning_tree_edge_ref = d['ref']
            spanning_tree_pos_node = d['pos_node']

            # For non-vsrc edges, positive current flows from the positive edge to the negative edge.
            # For vsrc edges, this direction is reversed. Therefore, the defined blue and green node
            # sets need swapped.
            if (d['info'].get_type() == ElementType.voltage_src):
                spanning_tree_pos_node = d['neg_node']

            self.qf[spanning_tree_edge_ref][spanning_tree_edge_ref] = 1

            for (i,j,e) in self.co_tree.edges(data=True):

                tree = self.spanning_tree.copy()

                parallel_edge = (u == i and v == j) or (u == j and v == i)

                if not parallel_edge:
                    tree.add_edge(i,j,e)

                loops = nx.cycle_basis(tree)
                loop_found = 0
                if len(loops) != 0:
                    if (u in loops[0] and v in loops[0]):
                        loop_found = 1

                if loop_found or parallel_edge:
                    if not parallel_edge:
                        tree.remove_edge(i,j)
                    tree.remove_edge(u,v)

                    co_tree_edge_ref = e['ref']
                    co_tree_pos_node = e['pos_node']

                    green_surface = nx.node_connected_component(tree, spanning_tree_pos_node)

                    if co_tree_pos_node in green_surface:
                        self.qf[spanning_tree_edge_ref][co_tree_edge_ref] = 1
                    else:
                        self.qf[spanning_tree_edge_ref][co_tree_edge_ref] = -1

    def get_bf_matrix(self):

        self.bf = np.zeros([self.num_edges, self.num_edges])

        for (u,v,d) in self.spanning_tree.edges(data=True):

            spanning_tree_edge_ref = d['ref']
            spanning_tree_pos_node = d['pos_node']
            spanning_tree_neg_node = d['neg_node']

            for (i,j,e) in self.co_tree.edges(data=True):
                tree = self.spanning_tree.copy()

                parallel_edge = (u == i and v == j) or (u == j and v == i)

                co_tree_edge_ref = e['ref']
                co_tree_pos_node = e['pos_node']
                co_tree_neg_node = e['neg_node']

                self.bf[co_tree_edge_ref][co_tree_edge_ref] = 1

                if parallel_edge:
                    if co_tree_pos_node == spanning_tree_pos_node:
                        self.bf[co_tree_edge_ref][spanning_tree_edge_ref] = -1
                    else:
                        self.bf[co_tree_edge_ref][spanning_tree_edge_ref] = 1
                else:
                    tree.add_edge(i,j,e)
                    loops = nx.cycle_basis(tree, co_tree_pos_node)
                    loop_found = len(loops) != 0

                    if loop_found:
                        if spanning_tree_pos_node in loops[0] and spanning_tree_neg_node in loops[0]:
                            spanning_tree_pos_node_index = loops[0].index(spanning_tree_pos_node)
                            spanning_tree_neg_node_index = loops[0].index(spanning_tree_neg_node)

                            co_tree_pos_node_index = loops[0].index(co_tree_pos_node)
                            co_tree_neg_node_index = loops[0].index(co_tree_neg_node)

                            spanning_tree_pos_traversal = ((loops[0][len(loops[0])-1] == spanning_tree_pos_node and
                                                            loops[0][0]               == spanning_tree_neg_node) or
                                                           (spanning_tree_pos_node_index == (spanning_tree_neg_node_index - 1)))

                            co_tree_pos_traversal = ((loops[0][len(loops[0])-1] == co_tree_pos_node and
                                                      loops[0][0]               == co_tree_neg_node) or
                                                     (co_tree_pos_node_index == (co_tree_neg_node_index - 1)))

                            if spanning_tree_pos_traversal == co_tree_pos_traversal:
                                self.bf[co_tree_edge_ref][spanning_tree_edge_ref] = 1
                            else:
                                self.bf[co_tree_edge_ref][spanning_tree_edge_ref] = -1

    #     VIN RD  ID  CJ  CT  CE  R
    # VIN -1  1   0   0   0   0   0
    # CJ  0   1   -1  -1  -1  0   0
    # CE  0   1   0   0   0   -1  -1
    #     VIN RD  ID  CJ  CT  CE  R
    # CT  0   0   0   -1  1   0   0
    # RD  -1  1   0   1   0   1   0
    # R   0   0   0   0   0   -1  1
    # ID  0   0   1   -1  0   0   0

    # With reference to the above matrix, the steps are as follows:
    # 1. The sys_var caps are CJ and CE. Process these rows in turn.
    #
    #    Initially create empty lists for each sys-var cap:
    #
    #         CJ CE
    #    CJ = [] []
    #    CE = [] []
    #
    #    For the sys-var element being processed, push a reference to itself in the correct list.
    #
    #         CJ   CE
    #    CJ = [CJ] []
    #    CE = []   [CE]
    #
    # 2. Loop through the columns in turn. When a non-sys var cap identifer is detected in the column name,
    #    find the corresponding row in the table. For example, CT is column 4 and row 3.
    #
    # 3. On the non-sys var cap row, find all the caps that are in a loop with this cap. Push a reference to the
    #    non-sys var cap into the list of the sys var cap. For example, consider the CT row:
    #
    #    CJ = [CJ, CT] []
    #
    # 4. In addition to the above, similar lists are required which specify the sign of the component and are equal to
    #    the corresponding Qm elements x -1.
    #
    #    CJ(sign) = [1, 1][]

    # The qm_eval_matrix is a 3D matrix of capacitor values that when multiplied by the capacitor values gives the Qm matrix.
    #
    # For the example above, the matrix would equal:
    #
    #    CT CE CJ  CT CE CJ
    # [ [1  0  1] [0  0  0]
    #   [0  0  0] [0  1  0] ]
    #
    # Therefore:
    #
    # Qm = [CT+CJ 0 ]
    #      [0     CE]
    #
    # The tm_eval_matrix matrix is a matrix which multipled by the edge values gives the Tm matrix.
    #
    # For the example above, the matrix would equal:
    #
    #   VIN RD  ID  CJ  CT  CE  R
    # [ 0   1   -1  0   0   0   0 ]
    # [ 0   1   0   0   0   -1  -1]
    #
    # Therefore:
    #
    # Tm = [iRD - iD]
    #      [iRD - iR]
    #
    # The tm_eval_cap_matrix is used to calculate Cn * dV values which occur when there is a capcitive/voltage loop in the netowrk.
    # Consider the Qf and Bf matrices shown below.
    #
    #     VS  C2  L1   C1  C3  R2
    # C2  0   1   0    0   1   -1
    # C1  0   0  -1    1   1   0
    #     VS  C2  L1   C1  C3  R2
    # L1  1   0   1    1   0   0
    # C3  1  -1   0   -1   1   0
    # R2 -1   1   0    0   0   1
    #
    # The system variable capacitors (C1 and C2) are both in the same basic cut as the non system variable capacitor, C3. To dermine the
    # current through C3, the voltage across it must be determined as a function of the system variables and independant voltage sources.
    # From the Bf matrix row:
    #
    # VC3 = VC1 + VC2 - VS
    #
    # Therefore, dVC3 = d (VC1 + VC2 - VS) and iC3 = C3 * dVC3 = C3 * (dVC1 + dVC2 - dVS).
    #
    # The (C3 * dVC1) and (C3 * dVC2) terms remain as part of the Qm matrix. However, the (C3 * dVS) term is part of the Tm matrix row.
    #
    # Since each basic cut can only contain one co-tree edge, there can only ever be one non system variable capacitor to consider. Therefore,
    # the tm_eval_cap_matrix is generated for every system variable capacitor row and contains a column for every capacitor in the network.
    # As with the tm_eval_matrix matrix, the tm_eval_cap_matrix, when mutlipled by the capacitor matrix, gives the capacitive value to multiply
    # each voltage source derivative by. For the example above, the matrix would equal:
    #
    #  C2  C1  C3
    # [0   0   1]
    # [0   0   1]
    #
    # The above problem could have been solved by placing the value of C3 into the VS column in the tm_eval_matrix. However, since capacitor
    # values can by dynamic, this value would need to be updated on every iteration which is slower than updating the capacitor matrix and
    # multiplying by the static eval matrix.
    #
    # Note that the tm_eval_matrix is always 0 for voltage sources.
    def get_dyf(self, eval_tree, is_sys_var, is_non_sys_var, get_edge_type_ref, eval_matrix, alt_matrix, am_eval_matrix, bm_eval_matrix, bm_eval_type_matrix):

        for (u,v,d) in eval_tree.edges(data=True):

            eval_tree_edge_ref         = d['ref']
            eval_tree_edge_sys_var_ref = self.get_sys_var_ref(eval_tree_edge_ref)

            if is_sys_var(eval_tree_edge_ref):

                # cap_ref or ind_ref
                type_ref = get_edge_type_ref(eval_tree_edge_ref)

                am_eval_matrix[eval_tree_edge_sys_var_ref][eval_tree_edge_sys_var_ref][type_ref] = 1

                # Sys var cap, identify all non sys var components also in this loop
                for j in range(self.num_edges):

                    # In the following code:
                    # Index j is the non-zero edge reference from the current spanning-tree qf row.
                    # Index z is the non-zero edge reference of qf row j (i.e. it is the index into the bf matrix).

                    # if a non-sys var cap, determine element voltage as a function of sys vars
                    if is_non_sys_var(j) and eval_matrix[eval_tree_edge_ref][j] != 0:

                        non_sys_var_polarity = eval_matrix[eval_tree_edge_ref][j]
                        non_sys_var_type_ref = get_edge_type_ref(j)

                        # If a non-sys var cap is identified that is part of the loop that contains the sys var cap,
                        # determine the voltage across the non sys cap expressed as a linear combination of the system variables.

                        # Get the Qm matrix coefficients and add to list
                        for z in range(self.num_edges):

                            alt_matrix_element = alt_matrix[j][z]

                            if alt_matrix_element != 0:

                                if is_sys_var(z):

                                    # Push a reference to edge j into list z.
                                    sys_var_ref = self.get_sys_var_ref(z)

                                    polarity = alt_matrix_element * non_sys_var_polarity * -1

                                    am_eval_matrix[eval_tree_edge_sys_var_ref][sys_var_ref][non_sys_var_type_ref] = polarity

                                elif z != j:
                                    # Find all components that are non-sys var caps. This should only be independant voltage sources.
                                    polarity = alt_matrix_element * non_sys_var_polarity

                                    bm_eval_type_matrix[eval_tree_edge_sys_var_ref][non_sys_var_type_ref] = polarity

                    # else the element in the loop with the sys-var cap is not a capacitor i.e. current source
                    elif eval_matrix[eval_tree_edge_ref][j] != 0 and j != eval_tree_edge_ref:
                        polarity = eval_matrix[eval_tree_edge_ref][j] * -1

                        bm_eval_matrix[eval_tree_edge_sys_var_ref][j] = polarity

        dict = {'am'     : am_eval_matrix, \
                'bm'     : bm_eval_matrix, \
                'bm_type': bm_eval_type_matrix}

        return dict

    # This function returns a vprobe string that identifies all system variable voltages and voltage sources
    # that a traversed between a two specified nodes.
    # A CSV string is returned of length num_components (circuit edges).
    # The returned string has a non-zero value (1 or -1 depending on polarity) if the edge is traversed.
    # The edge index (assigned when reading the netlist) defines the position of the 1 or -1.
    # Non-zero values should only appear in locations whose index is a system variable or voltage source.
    def get_vprobe(self, pos_node, neg_node):

        vprobe = np.zeros(self.num_edges)

        for i in range(self.num_edges):

            path = nx.shortest_path(self.spanning_tree, source=pos_node, target=neg_node)

            # Determine if the path from source to target is traversed in the order of pos_node -> neg_node
            # For example, if pos_node = 10 and neg_node = 4 and the path is [10, 0, 8, 4] then traversal is
            # negative since the path moves from 4->10
            pos_node_index = path.index(pos_node)
            neg_node_index = path.index(neg_node)

            spanning_tree_pos_traversal = ((path[len(path)-1] == pos_node and
                                            path[0]           == neg_node) or
                                           (pos_node_index == (neg_node_index - 1)))

            # A single element path is found when the pos_node/neg_node pair are connected by a single element
            # This must be either a capacitor or voltage source
            single_element_path = (len(path) == 2)

            # Determine if when moving terms to RHS of equations their magnitude is positive or negative
            # For example, if:
            # V(+-) + VA + VB - VC = 0 and V(+-) was traversed positively, then:
            # V(+-) = -VA - VB + VC
            # When V(+-) is a single edge, the traversal is always positive
            traversal_mul = 1
            if spanning_tree_pos_traversal == True and not single_element_path:
                traversal_mul = -1

            for i in range(len(path)-1):
                d = self.spanning_tree.get_edge_data(path[i], path[i+1])

                edge_ref      = d['ref']
                pos_node_edge = d['pos_node']

                # determine positive/negative traversal
                if int(pos_node_edge) == int(path[i]):
                    vprobe[edge_ref] = 1 * traversal_mul
                else:
                    vprobe[edge_ref] = -1 * traversal_mul

        return vprobe

    def init_circuit(self):
        self.consistency_check()
        self.get_spanning_tree()
        self.get_co_tree()
        self.set_sys_var_ref()
        self.set_dep_var_ref()
        self.set_src_edge_ref()
        self.get_qf_matrix()
        self.get_bf_matrix()

        self.tm_eval_matrix     = np.zeros([self.num_sys_var_cap, self.num_edges])
        self.qm_eval_matrix     = np.zeros([self.num_sys_var_cap, self.num_sys_var_cap, self.num_caps])
        self.tm_eval_cap_matrix = np.zeros([self.num_sys_var_cap, self.num_caps])
        eval_matrices = self.get_dyf(self.spanning_tree, \
                                     self.is_sys_var_cap, \
                                     self.is_non_sys_var_cap, \
                                     self.get_edge_cap_ref, \
                                     self.qf, \
                                     self.bf, \
                                     self.qm_eval_matrix, \
                                     self.tm_eval_matrix, \
                                     self.tm_eval_cap_matrix)

        self.qm_eval_matrix     = eval_matrices['am']
        self.tm_eval_matrix     = eval_matrices['bm']
        self.tm_eval_cap_matrix = eval_matrices['bm_type']

        self.zm_eval_matrix     = np.zeros([self.num_sys_var_ind, self.num_edges])
        self.ym_eval_matrix     = np.zeros([self.num_sys_var_ind, self.num_sys_var_ind, self.num_inductors])
        self.zm_eval_ind_matrix = np.zeros([self.num_sys_var_ind, self.num_inductors])
        eval_matrices = self.get_dyf(self.co_tree, \
                                     self.is_sys_var_ind, \
                                     self.is_non_sys_var_ind, \
                                     self.get_edge_ind_ref, \
                                     self.bf, \
                                     self.qf, \
                                     self.ym_eval_matrix, \
                                     self.zm_eval_matrix, \
                                     self.zm_eval_ind_matrix)

        self.ym_eval_matrix     = eval_matrices['am']
        self.zm_eval_matrix     = eval_matrices['bm']
        self.zm_eval_cap_matrix = eval_matrices['bm_type']

        self.set_sys_var_cap_list()
        self.set_non_sys_var_cap_list()
        self.set_sys_var_ind_list()
        self.set_non_sys_var_ind_list()
        self.set_co_tree_res_list()
        self.set_spanning_tree_res_list()
        self.create_edge_dict()

    def get_qm_matrix(self, cap_values):

        var_1 = self.num_sys_var_cap * self.num_sys_var_cap

        # unpack the self.qm_eval_matrix matrix:
        # original:
        #  [[a00, b00, c00, ..., x00], [a01, b01, c01, ..., x01],..., [a0n, b0n, c0n, ..., x0n],
        #   [a10, b10, c10, ..., x10], [a11, b11, c11, ..., x11],..., [a1n, b1n, c1n, ..., x1n],
        #   ...
        #   [am0, bm0, cm0, ..., xm0], [am1, bm1, cm1, ..., xm1],..., [amn, bmn, cmn, ..., xmn]]
        #
        # unpacked:
        #  [[a00, b00, c00, ..., x00]
        #   [a01, b01, c01, ..., x01]
        #   ...
        #   [a0n, b0n, c0n, ..., x0n]
        #   [a10, b10, c10, ..., x10]
        #   [a11, b11, c11, ..., x11]
        #   [a1n, b1n, c1n, ..., x1n]
        #   ...
        #   [am0, bm0, cm0, ..., xm0]
        #   [am1, bm1, cm1, ..., xm1]
        #   [amn, bmn, cmn, ..., xmn]]
        #
        # Each row of the unpacked array is then multiplied by self.cap_values_f.
        # The result of the multiplication is then reshaped to (self.num_sys_var_cap, self.num_sys_var_cap)
        #
        # The above is equivalent to:
        # for i in range(self.num_sys_var_cap):
        #    for j in range(self.num_sys_var_cap):
        #       self.qm[i][j] = self.qm_eval_matrix[i][j].dot(self.cap_values_f)

        a = self.qm_eval_matrix.transpose(0,1,2).reshape(var_1,-1)
        b = a.dot(cap_values)
        qm_matrix = b.reshape(self.num_sys_var_cap, self.num_sys_var_cap)

        return qm_matrix

    def get_ym_matrix(self, ind_values):

        var_1 = self.num_sys_var_ind * self.num_sys_var_ind

        # unpack the self.qm_eval_matrix matrix:
        # original:
        #  [[a00, b00, c00, ..., x00], [a01, b01, c01, ..., x01],..., [a0n, b0n, c0n, ..., x0n],
        #   [a10, b10, c10, ..., x10], [a11, b11, c11, ..., x11],..., [a1n, b1n, c1n, ..., x1n],
        #   ...
        #   [am0, bm0, cm0, ..., xm0], [am1, bm1, cm1, ..., xm1],..., [amn, bmn, cmn, ..., xmn]]
        #
        # unpacked:
        #  [[a00, b00, c00, ..., x00]
        #   [a01, b01, c01, ..., x01]
        #   ...
        #   [a0n, b0n, c0n, ..., x0n]
        #   [a10, b10, c10, ..., x10]
        #   [a11, b11, c11, ..., x11]
        #   [a1n, b1n, c1n, ..., x1n]
        #   ...
        #   [am0, bm0, cm0, ..., xm0]
        #   [am1, bm1, cm1, ..., xm1]
        #   [amn, bmn, cmn, ..., xmn]]
        #
        # Each row of the unpacked array is then multiplied by self.cap_values_f.
        # The result of the multiplication is then reshaped to (self.num_sys_var_cap, self.num_sys_var_cap)
        #
        # The above is equivalent to:
        # for i in range(self.num_sys_var_cap):
        #    for j in range(self.num_sys_var_cap):
        #       self.qm[i][j] = self.qm_eval_matrix[i][j].dot(self.cap_values_f)

        a = self.ym_eval_matrix.transpose(0,1,2).reshape(var_1,-1)
        b = a.dot(ind_values)
        ym_matrix = b.reshape(self.num_sys_var_ind, self.num_sys_var_ind)

        return ym_matrix

    def print_ind_dy(self):

        for i in range(self.num_sys_var_ind):

            row_str = ""
            for j in range(len(self.ind_dy_b[i])):
                edge_info = self.ind_dy_b[i][j]

                element_str = "["
                for k in range(len(edge_info)):
                    if self.ind_dy_b_sign[i][j][k] == 1:
                        sign = '+'
                    else:
                        sign = '-'

                    element_str += sign + edge_info[k].get_instance() + " "
                element_str += "]"
                row_str += element_str
            print(row_str)

    def print_ind_dy_t(self):

        for i in range(self.num_sys_var_ind):

            row_str = ""

            for j in range(len(self.ind_dy_t[i])):
                edge_info = self.ind_dy_t[i][j]

                # Check if the edge is used.
                # An empty list entry [] evaluates to False
                if edge_info:

                    if self.ind_dy_t_sign[i][j][0] == 1:
                        sign = '+'
                    else:
                        sign = '-'

                    # Determine if edge j is a voltage source
                    if self.is_isrc(j):
                        row_str += sign + '(' + edge_info[0].get_instance() + " * d" + self.get_edge_info(j).get_instance() + ')'
                    else:
                        row_str += sign + edge_info[0].get_instance()
            print(row_str)

    def print_edges(self):

        #for e in self.G.edges_iter(keys=True, data=True):
         #   print(e)

        for (u,v,d) in self.G.edges(data=True):
            #print(u)
            #print(v)
            #print(d)

            pos_node = d['pos_node']
            neg_node = d['neg_node']
            edge     = d['info']
            edge_ref = d['ref']
            sys_var_ref = d['sys_var_ref']

            print ("--")
            print (edge_ref)
            print (sys_var_ref)
            print (pos_node)
            print (neg_node)
            print (edge.get_type())
            print (edge.get_instance())
            print (edge)
            #print(edge.get_value(self, self.t))

    def create_edge_dict(self):

        edge_dict = [None] * self.num_edges

        for (u,v,d) in self.spanning_tree.edges(data=True):

            edge = d['info']

            if edge.get_type() == ElementType.capacitor:
                edge_ref = d['ref']
                cap_ref  = d['cap_ref']

                edge_dict_entry = {}
                edge_dict_entry['cap_ref'] = cap_ref
                edge_dict[edge_ref] = edge_dict_entry

            elif edge.get_type() == ElementType.inductor:
                edge_ref = d['ref']
                ind_ref  = d['ind_ref']

                edge_dict_entry = {}
                edge_dict_entry['ind_ref'] = ind_ref
                edge_dict[edge_ref] = edge_dict_entry

        for (u,v,d) in self.co_tree.edges(data=True):

            edge = d['info']

            if edge.get_type() == ElementType.capacitor:
                edge_ref = d['ref']
                cap_ref  = d['cap_ref']

                edge_dict_entry = {}
                edge_dict_entry['cap_ref'] = cap_ref
                edge_dict[edge_ref] = edge_dict_entry

            elif edge.get_type() == ElementType.inductor:
                edge_ref = d['ref']
                ind_ref  = d['ind_ref']

                edge_dict_entry = {}
                edge_dict_entry['ind_ref'] = ind_ref
                edge_dict[edge_ref] = edge_dict_entry

        return edge_dict

    def get_edge_dict(self):

        edge_dict = [None] * self.num_edges

        for (u,v,d) in self.G.edges(data=True):

            edge     = d['info']
            edge_ref = d['ref']
            instance = edge.get_instance()
            value    = edge.get_value([], [], 0)

            if self.is_sys_var_cap(edge_ref):
                cap_ref  = d['cap_ref']

             #   print (edge)
            #print (edge.get_instance())
            #print (edge_ref)
